# -*- coding: utf-8 -*-
"""
    bromelia.constants.etsi_3gpp.ts_129_214
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    This module contains constants defined in ETSI TS 129 214.
    
    :copyright: (c) 2020-present Henrique Marques Ribeiro.
    :license: MIT, see LICENSE for more details.
"""

from ..._internal_utils import convert_to_4_bytes


#: Diameter AVPs
ABORT_CAUSE_AVP_CODE = convert_to_4_bytes(500)
ACCESS_NETWORK_CHARGING_ADDRESS_AVP_CODE = convert_to_4_bytes(501)
ACCESS_NETWORK_CHARGING_IDENTIFIER_VALUE_AVP_CODE = convert_to_4_bytes(503)
AF_APPLICATION_IDENTIFIER_AVP_CODE = convert_to_4_bytes(504)
AF_CHARGING_IDENTIFIER_AVP_CODE = convert_to_4_bytes(505)
FLOW_DESCRIPTION_AVP_CODE = convert_to_4_bytes(507)
FLOW_NUMBER_AVP_CODE = convert_to_4_bytes(509)
FLOW_STATUS_AVP_CODE = convert_to_4_bytes(511)
FLOW_USAGE_AVP_CODE = convert_to_4_bytes(512)
SPECIFIC_ACTION_AVP_CODE = convert_to_4_bytes(513)
MAX_REQUESTED_BANDWIDTH_DL_AVP_CODE = convert_to_4_bytes(515)
MAX_REQUESTED_BANDWIDTH_UL_AVP_CODE = convert_to_4_bytes(516)
MEDIA_COMPONENT_DESCRIPTION_AVP_CODE = convert_to_4_bytes(517)
MEDIA_COMPONENT_NUMBER_AVP_CODE = convert_to_4_bytes(518) 
MEDIA_SUB_COMPONENT_AVP_CODE = convert_to_4_bytes(519)
MEDIA_TYPE_AVP_CODE = convert_to_4_bytes(520)
SERVICE_INFO_STATUS_AVP_CODE = convert_to_4_bytes(527)

#: List of Abort-Cause AVP values.
#: For more information, please refer to Section 5.3.1 of 
#: ETSI TS 129 214 V15.4.0 (2018-07).
ABORT_CAUSE_BEARER_RELEASED = convert_to_4_bytes(0)
ABORT_CAUSE_INSUFFICIENT_SERVER_RESOURCES = convert_to_4_bytes(1)
ABORT_CAUSE_INSUFFICIENT_BEARER_RESOURCES = convert_to_4_bytes(2)
ABORT_CAUSE_PS_TO_CS_HANDOVER = convert_to_4_bytes(3)
ABORT_CAUSE_SPONSORED_DATA_CONNECTIVITY_DISALLOWED = convert_to_4_bytes(4)

#: List of Flow-Status AVP values.
#: For more information, please refer to Section 5.3.11 of 
#: ETSI TS 129 214 V15.4.0 (2018-07).
FLOW_STATUS_ENABLED_UPLINK = convert_to_4_bytes(0)
FLOW_STATUS_ENABLED_DOWNLINK = convert_to_4_bytes(1)
FLOW_STATUS_ENABLED = convert_to_4_bytes(2)
FLOW_STATUS_DISABLED = convert_to_4_bytes(3)
FLOW_STATUS_REMOVED = convert_to_4_bytes(4)

#: List of Flow-Usage AVP values.
#: For more information, please refer to Section 5.3.12 of 
#: ETSI TS 129 214 V15.4.0 (2018-07).
FLOW_USAGE_NO_INFORMATION = convert_to_4_bytes(0)
FLOW_USAGE_RTCP = convert_to_4_bytes(1)
FLOW_USAGE_AF_SIGNALLING = convert_to_4_bytes(2)

#: List of Specific-Action AVP values.
#: For more information, please refer to Section 5.3.13 of 
#: ETSI TS 129 214 V15.4.0 (2018-07).
SPECIFIC_ACTION_CHARGING_CORRELATION_EXCHANGE = convert_to_4_bytes(1)
SPECIFIC_ACTION_INDICATION_OF_LOSS_OF_BEARER = convert_to_4_bytes(2)
SPECIFIC_ACTION_INDICATION_OF_RECOVERY_OF_BEARER = convert_to_4_bytes(3)
SPECIFIC_ACTION_INDICATION_OF_RELEASE_OF_BEARER = convert_to_4_bytes(4)
SPECIFIC_ACTION_IP_CAN_CHANGE = convert_to_4_bytes(6)
SPECIFIC_ACTION_INDICATION_OF_OUT_OF_CREDIT = convert_to_4_bytes(7)
SPECIFIC_ACTION_INDICATION_OF_SUCCESSFUL_RESOURCES_ALLOCATION = convert_to_4_bytes(8)
SPECIFIC_ACTION_INDICATION_OF_FAILED_RESOURCES_ALLOCATION = convert_to_4_bytes(9)
SPECIFIC_ACTION_INDICATION_OF_LIMITED_PCC_DEPLOYMENT = convert_to_4_bytes(10)
SPECIFIC_ACTION_USAGE_REPORT = convert_to_4_bytes(11)
SPECIFIC_ACTION_ACCESS_NETWORK_INFO_REPORT = convert_to_4_bytes(12)
SPECIFIC_ACTION_INDICATION_OF_RECOVERY_FROM_LIMITED_PCC_DEPLOYMENT = convert_to_4_bytes(13)
SPECIFIC_ACTION_INDICATION_OF_ACCESS_NETWORK_INFO_REPORTING_FAILURE = convert_to_4_bytes(14)
SPECIFIC_ACTION_INDICATION_OF_TRANSFER_POLICY_EXPIRED = convert_to_4_bytes(15)
SPECIFIC_ACTION_PLMN_CHANGE = convert_to_4_bytes(16)

#: List of Media-Type AVP values.
#: For more information, please refer to Section 5.3.19 of 
#: ETSI TS 129 214 V15.4.0 (2018-07).
MEDIA_TYPE_AUDIO = convert_to_4_bytes(0)
MEDIA_TYPE_VIDEO = convert_to_4_bytes(1)
MEDIA_TYPE_DATA = convert_to_4_bytes(2)
MEDIA_TYPE_APPLICATION = convert_to_4_bytes(3)
MEDIA_TYPE_CONTROL = convert_to_4_bytes(4)
MEDIA_TYPE_TEXT = convert_to_4_bytes(5)
MEDIA_TYPE_MESSAGE = convert_to_4_bytes(6)
MEDIA_TYPE_OTHER = convert_to_4_bytes(7)

#: List of Service-Info-Status AVP values.
#: For more information, please refer to Section 5.3.25 of 
#: ETSI TS 129 214 V15.4.0 (2018-07).
SERVICE_INFO_STATUS_FINAL_SERVICE_INFORMATION = convert_to_4_bytes(0)
SERVICE_INFO_STATUS_PRELIMINARY_SERVICE_INFORMATION = convert_to_4_bytes(1)



# -*- coding: utf-8 -*-
"""
    bromelia.etsi_3gpp.ts_129_329
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    This module contains Diameter AVP classes defined in ETSI TS 129 329.
    
    :copyright: (c) 2020-present Henrique Marques Ribeiro.
    :license: MIT, see LICENSE for more details.
"""

from ...base import DiameterAVP
from ...constants.etsi_3gpp.ts_129_329 import *
from ...types import *
from ...utils import decode_from_tbcd, encode_to_tbcd


class MsisdnAVP(DiameterAVP, OctetStringType):
    """Implementation of MSISDN AVP in Section 6.3.2 of 
    ETSI TS 129 329 V15.1.0 (2018-07).

    The MSISDN AVP (AVP Code 701) is of type OctetString.
    """
    code = MSISDN_AVP_CODE
    vendor_id = VENDOR_ID_3GPP

    def __init__(self, data):
        DiameterAVP.__init__(self, 
                             MsisdnAVP.code,
                             MsisdnAVP.vendor_id)
        DiameterAVP.set_mandatory_bit(self, True)
        DiameterAVP.set_vendor_id_bit(self, True)
        OctetStringType.__init__(self, data=self.encode(data), vendor_id=VENDOR_ID_3GPP)


    def encode(self, data):
        if isinstance(data, int):
            return bytes.fromhex(encode_to_tbcd(data))
        
        elif isinstance(data, str):
            return bytes.fromhex(encode_to_tbcd(int(data)))

        elif isinstance(data, bytes):
            return data


    def decode(self):
        return decode_from_tbcd(self.data)


# -*- coding: utf-8 -*-
"""
    bromelia.lib.ietf_rfc6733.messages
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    This module contains the Diameter protocol base messages.

    :copyright: (c) 2020-present Henrique Marques Ribeiro.
    :license: MIT, see LICENSE for more details.
"""

import platform
import socket

from ...avps import *
from ...base import DiameterRequest, DiameterAnswer
from ...constants import *
from ...exceptions import DiameterMessageError


class AbortSessionAnswer(DiameterAnswer):
    """Implementation of Abort-Session-Answer (ASA) in Section 8.5.2 of 
    IETF RFC 6733.

    The Abort-Session-Answer is indicated by the Command Code 274 and the 
    Command Flags' 'R' bit cleared.

    Usage::

        >>> from bromelia.messages import ASA
        >>> from bromelia import DIAMETER_APPLICATION_Rx
        >>> asa = ASA()
        >>> asa.header.application_id = DIAMETER_APPLICATION_Rx
        >>> asa
        <Diameter Message: 274 [ASA] PXY 3GPP Rx, 4 AVP(s)>
    """

    mandatory = {
                    "session_id": SessionIdAVP,
                    "result_code": ResultCodeAVP,
                    "origin_host": OriginHostAVP,
                    "origin_realm": OriginRealmAVP,
    }
    optionals = {
                    "user_name": UserNameAVP,
                    "origin_state_id": OriginStateIdAVP,
                    "error_message": ErrorMessageAVP,
                    "error_reporting_host": ErrorReportingHostAVP,
                    "failed_avp": FailedAvpAVP,
                    "redirect_host": RedirectHostAVP,
                    "redirect_host_usage": RedirectHostUsageAVP,
                    "redirect_max_cache_time": RedirectMaxCacheTimeAVP,
                    "proxy_info": ProxyInfoAVP
    }

    def __init__(self, 
                session_id=platform.node(), 
                result_code=DIAMETER_SUCCESS,
                origin_host=platform.node(), 
                origin_realm=socket.getfqdn(), 
                user_name=None,
                origin_state_id=None,
                error_message=None,
                error_reporting_host=None,
                failed_avp=None,
                redirect_host=None,
                redirect_host_usage=None,
                redirect_max_cache_time=None,
                proxy_info=None,
                **kwargs):
        
        DiameterAnswer.__init__(self, 
                                command_code=ABORT_SESSION_MESSAGE, 
                                application_id=None)

        DiameterAnswer._load(self, locals())


class AbortSessionRequest(DiameterRequest):
    """Implementation of Abort-Session-Request (ASR) in Section 8.5.1 of 
    IETF RFC 6733.

    The Abort-Session-Request is indicated by the Command Code 274 and the 
    Command Flags' 'R' bit set.

    Usage::

        >>> from bromelia.messages import ASR
        >>> from bromelia import DIAMETER_APPLICATION_Rx
        >>> asr_avps = {
        ...     "auth_application_id": DIAMETER_APPLICATION_Rx,
        ...     "destination_realm": "example.com",
        ...     "destination_host": "host.example.com"
        ... }
        >>> asr = ASR(**asr_avps)
        >>> asr
        <Diameter Message: 274 [ASR] REQ, PXY 3GPP Rx, 6 AVP(s)>
    """

    mandatory = { 
                    "session_id": SessionIdAVP,
                    "origin_host": OriginHostAVP,
                    "origin_realm": OriginRealmAVP,
                    "destination_realm": DestinationRealmAVP,
                    "destination_host": DestinationHostAVP,
                    "auth_application_id": AuthApplicationIdAVP
    }
    optionals = { 
                    "user_name": UserNameAVP,
                    "origin_state_id": OriginStateIdAVP,
                    "proxy_info": ProxyInfoAVP,
                    "route_record": RouteRecordAVP
    }
    
    def __init__(self,
                session_id=platform.node(), 
                origin_host=platform.node(), 
                origin_realm=socket.getfqdn(), 
                destination_realm=socket.gethostbyname(platform.node()),
                destination_host=None,
                auth_application_id=None,
                user_name=None,
                origin_state_id=None,
                proxy_info=None,
                route_record=None,
                **kwargs):

        if not auth_application_id:
            raise DiameterMessageError("invalid auth_application_id value. "\
                                    "It needs to include a valid Auth "\
                                    "Application Id")

        DiameterRequest.__init__(self, 
                                 command_code=ABORT_SESSION_MESSAGE,
                                 application_id=auth_application_id)

        DiameterRequest._load(self, locals())


class CapabilitiesExchangeAnswer(DiameterAnswer):
    """Implementation of Capabilities-Exchange-Answer (CEA) in Section 5.3.2 
    of IETF RFC 6733.

    The Capabilities-Exchange-Answer is indicated by the Command Code 257 and
    the Command Flags' 'R' bit cleared.

    Usage::

        >>> from bromelia.messages import CEA
        >>> cea = CEA()
        >>> cea
        <Diameter Message: 257 [CEA], Default, 6 AVP(s)>
    """

    mandatory = {
                    "result_code": ResultCodeAVP,
                    "origin_host": OriginHostAVP,
                    "origin_realm": OriginRealmAVP,
                    "host_ip_address": HostIpAddressAVP,
                    "vendor_id": VendorIdAVP,
                    "product_name": ProductNameAVP
    }
    optionals = {
                    "origin_state_id": OriginStateIdAVP,
                    "error_message": ErrorMessageAVP,
                    "failed_avp": FailedAvpAVP,
                    "supported_vendor_id": SupportedVendorIdAVP,
                    "auth_application_id": AuthApplicationIdAVP,
                    "inband_security_id": InbandSecurityIdAVP,
                    "acct_application_id": AcctApplicationIdAVP,
                    "vendor_specific_application_id": VendorSpecificApplicationIdAVP,
                    "firmware_revision": FirmwareRevisionAVP
    }

    def __init__(self, 
                result_code=DIAMETER_SUCCESS,
                origin_host=platform.node(), 
                origin_realm=socket.getfqdn(), 
                host_ip_address=socket.gethostbyname(platform.node()),
                vendor_id=0,
                product_name=PRODUCT_NAME,
                origin_state_id=None,
                error_message=None,
                failed_avp=None,
                supported_vendor_id=None,
                auth_application_id=DIAMETER_APPLICATION_DEFAULT,
                inband_security_id=None,
                acct_application_id=None,
                vendor_specific_application_id=None, 
                firmware_revision=None,
                **kwargs):

        DiameterAnswer.__init__(self,
                                command_code=CAPABILITIES_EXCHANGE_MESSAGE,
                                application_id=DIAMETER_APPLICATION_DEFAULT)

        DiameterAnswer._load(self, locals())


class CapabilitiesExchangeRequest(DiameterRequest):
    """Implementation of Capabilities-Exchange-Request (CER) in Section 5.3.1 
    of IETF RFC 6733.

    The Capabilities-Exchange-Request is indicated by the Command Code 257 and
    the Command Flags' 'R' bit set.

    Usage::

        >>> from bromelia.messages import CER
        >>> cer = CER()
        >>> cer
        <Diameter Message: 257 [CER], REQ Default, 6 AVP(s)>
    """

    mandatory = { 
                    "origin_host": OriginHostAVP,
                    "origin_realm": OriginRealmAVP,
                    "host_ip_address": HostIpAddressAVP,
                    "vendor_id": VendorIdAVP,
                    "product_name": ProductNameAVP
    }
    optionals = { 
                    "origin_state_id": OriginStateIdAVP,
                    "supported_vendor_id": SupportedVendorIdAVP,
                    "auth_application_id": AuthApplicationIdAVP,
                    "inband_security_id": InbandSecurityIdAVP,
                    "acct_application_id": AcctApplicationIdAVP,
                    "vendor_specific_application_id": VendorSpecificApplicationIdAVP,
                    "firmware_revision": FirmwareRevisionAVP
    }
    
    def __init__(self, 
                origin_host=platform.node(), 
                origin_realm=socket.getfqdn(), 
                host_ip_address=socket.gethostbyname(platform.node()),
                vendor_id=VENDOR_ID_DEFAULT,
                product_name=PRODUCT_NAME,
                origin_state_id=None,
                supported_vendor_id=None,
                auth_application_id=None,
                inband_security_id=None,
                acct_application_id=None,
                vendor_specific_application_id=None, 
                firmware_revision=FIRMWARE_VERSION,
                **kwargs):

        DiameterRequest.__init__(self, 
                                command_code=CAPABILITIES_EXCHANGE_MESSAGE,
                                application_id=DIAMETER_APPLICATION_DEFAULT)

        DiameterRequest._load(self, locals())


class DeviceWatchdogAnswer(DiameterAnswer):
    """Implementation of Device-Watchdog-Answer (DWA) in Section 5.5.2 
    of IETF RFC 6733.

    The Device-Watchdog-Answer is indicated by the Command Code 280 and the
    Command Flags' 'R' bit cleared.

    Usage::

        >>> from bromelia.messages import DWA
        >>> dwa = DWA()
        >>> dwa
        <Diameter Message: 280 [DWA], Default, 3 AVP(s)>
    """

    mandatory = {
                    "result_code": ResultCodeAVP,
                    "origin_host": OriginHostAVP,
                    "origin_realm": OriginRealmAVP
    }
    optionals = {
                    "error_message": ErrorMessageAVP,
                    "failed_avp": FailedAvpAVP,
                    "origin_state_id": OriginStateIdAVP
    }

    def __init__(self,
                result_code=DIAMETER_SUCCESS,
                origin_host=platform.node(), 
                origin_realm=socket.getfqdn(),
                error_message=None,
                failed_avp=None,
                origin_state_id=None,
                **kwargs):

        DiameterAnswer.__init__(self, 
                                command_code=DEVICE_WATCHDOG_MESSAGE,
                                application_id=DIAMETER_APPLICATION_DEFAULT)

        DiameterAnswer._load(self, locals())


class DeviceWatchdogRequest(DiameterRequest):
    """Implementation of Device-Watchdog-Request (DWR) in Section 5.5.1 
    of IETF RFC 6733.

    The Device-Watchdog-Request is indicated by the Command Code 280 and the
    Command Flags' 'R' bit set.

    Usage::

        >>> from bromelia.messages import DWR
        >>> dwr = DWR()
        >>> dwr
        <Diameter Message: 280 [DWR], REQ Default, 2 AVP(s)>
    """

    mandatory = {
                    "origin_host": OriginHostAVP,
                    "origin_realm": OriginRealmAVP,
    }
    optionals = {
                    "origin_state_id": OriginStateIdAVP
    }
    
    def __init__(self,
                origin_host=platform.node(), 
                origin_realm=socket.getfqdn(), 
                origin_state_id=None,
                **kwargs):

        DiameterRequest.__init__(self, 
                                command_code=DEVICE_WATCHDOG_MESSAGE,
                                application_id=DIAMETER_APPLICATION_DEFAULT)

        DiameterRequest._load(self, locals())


class DisconnectPeerAnswer(DiameterAnswer):
    """Implementation of Disconnect-Peer-Answer (DPA) in Section 5.4.2 
    of IETF RFC 6733.

    The Disconnect-Peer-Answer is indicated by the Command Code 282 and
    the Command Flags' 'R' bit cleared.

    Usage::

        >>> from bromelia.messages import DPA
        >>> dpa = DPA()
        >>> dpa
        <Diameter Message: 282 [DPA], Default, 3 AVP(s)>
    """
    mandatory = {
                    "result_code": ResultCodeAVP,
                    "origin_host": OriginHostAVP,
                    "origin_realm": OriginRealmAVP,
    }
    optionals = {
                    "error_message": ErrorMessageAVP,
                    "failed_avp": FailedAvpAVP
    }

    def __init__(self,
                result_code=DIAMETER_SUCCESS,
                origin_host=platform.node(), 
                origin_realm=socket.getfqdn(), 
                error_message=None,
                failed_avp=None,
                **kwargs):

        DiameterAnswer.__init__(self, 
                                command_code=DISCONNECT_PEER_MESSAGE,
                                application_id=DIAMETER_APPLICATION_DEFAULT)

        DiameterAnswer._load(self, locals())


class DisconnectPeerRequest(DiameterRequest):
    """Implementation of Disconnect-Peer-Request (DPR) in Section 5.4.1 
    of IETF RFC 6733.

    The Disconnect-Peer-Request is indicated by the Command Code 282 and
    the Command Flags' 'R' bit set.

    Usage::

        >>> from bromelia.messages import DPR
        >>> dpr = DPR()
        >>> dpr
        <Diameter Message: 282 [DPR], REQ Default, 3 AVP(s)>
    """

    mandatory = {
                    "origin_host": OriginHostAVP,
                    "origin_realm": OriginRealmAVP,
                    "disconnect_cause": DisconnectCauseAVP,
    }
    optionals = {}

    def __init__(self,
                origin_host=platform.node(), 
                origin_realm=socket.getfqdn(), 
                disconnect_cause=DISCONNECT_CAUSE_REBOOTING,
                **kwargs):

        DiameterRequest.__init__(self, 
                                command_code=DISCONNECT_PEER_MESSAGE,
                                application_id=DIAMETER_APPLICATION_DEFAULT)

        DiameterRequest._load(self, locals())


class ReAuthAnswer(DiameterAnswer):
    """Implementation of Re-Auth-Answer (RAA) in Section 8.3.2 of 
    IETF RFC 6733.

    The Re-Auth-Answer is indicated by the Command Code 258 and the Command 
    Flags' 'R' bit cleared.

    Usage::

        >>> from bromelia.messages import RAA
        >>> from bromelia import DIAMETER_APPLICATION_Rx
        >>> raa = RAA()
        >>> raa.header.application_id = DIAMETER_APPLICATION_Rx
        >>> raa
        <Diameter Message: 258 [RAA] PXY 3GPP Rx, 4 AVP(s)>
    """

    mandatory = {
                    "session_id": SessionIdAVP,
                    "result_code": ResultCodeAVP,
                    "origin_host": OriginHostAVP,
                    "origin_realm": OriginRealmAVP,
    }
    optionals = {
                    "user_name": UserNameAVP,
                    "origin_state_id": OriginStateIdAVP,
                    "error_message": ErrorMessageAVP,
                    "error_reporting_host": ErrorReportingHostAVP,
                    "failed_avp": FailedAvpAVP,
                    "redirect_host": RedirectHostAVP,
                    "redirect_host_usage": RedirectHostUsageAVP,
                    "redirect_max_cache_time": RedirectMaxCacheTimeAVP,
                    "proxy_info": ProxyInfoAVP
    }

    def __init__(self, 
                session_id=platform.node(), 
                result_code=DIAMETER_SUCCESS,
                origin_host=platform.node(), 
                origin_realm=socket.getfqdn(), 
                user_name=None,
                origin_state_id=None,
                error_message=None,
                error_reporting_host=None,
                failed_avp=None,
                redirect_host=None,
                redirect_host_usage=None,
                redirect_max_cache_time=None,
                proxy_info=None,
                **kwargs):
        
        DiameterAnswer.__init__(self, 
                                command_code=RE_AUTH_MESSAGE, 
                                application_id=None)

        DiameterAnswer._load(self, locals())


class ReAuthRequest(DiameterRequest):
    """Implementation of Re-Auth-Request (RAR) in Section 8.3.1 of 
    IETF RFC 6733.

    The Re-Auth-Request is indicated by the Command Code 258 and the 
    Command Flags' 'R' bit set.

    Usage::

        >>> from bromelia.messages import RAR
        >>> from bromelia import DIAMETER_APPLICATION_Gx
        >>> from bromelia import AUTH_REQUEST_TYPE_AUTHENTICATE_ONLY
        >>> rar_avps = {
        ...     "auth_application_id": DIAMETER_APPLICATION_Gx,
        ...     "destination_realm": "example.com",
        ...     "destination_host": "host.example.com",
        ...     "re_auth_request_type": AUTH_REQUEST_TYPE_AUTHENTICATE_ONLY
        ... }
        >>> rar = RAR(**rar_avps)
        >>> rar
        <Diameter Message: 258 [RAR] REQ, PXY 3GPP Gx, 7 AVP(s)>
    """

    mandatory = { 
                    "session_id": SessionIdAVP,
                    "origin_host": OriginHostAVP,
                    "origin_realm": OriginRealmAVP,
                    "destination_realm": DestinationRealmAVP,
                    "destination_host": DestinationHostAVP,
                    "auth_application_id": AuthApplicationIdAVP,
                    "re_auth_request_type": ReAuthRequestTypeAVP
    }
    optionals = { 
                    "user_name": UserNameAVP,
                    "origin_state_id": OriginStateIdAVP,
                    "proxy_info": ProxyInfoAVP,
                    "route_record": RouteRecordAVP
    }
    
    def __init__(self,
                session_id=platform.node(), 
                origin_host=platform.node(), 
                origin_realm=socket.getfqdn(), 
                destination_realm=socket.gethostbyname(platform.node()),
                destination_host=None,
                auth_application_id=None,
                re_auth_request_type=None,
                user_name=None,
                origin_state_id=None,
                proxy_info=None,
                route_record=None,
                **kwargs):

        if not auth_application_id:
            raise DiameterMessageError("invalid auth_application_id value. "\
                                    "It needs to include a valid Auth "\
                                    "Application Id")

        DiameterRequest.__init__(self, 
                                 command_code=RE_AUTH_MESSAGE, 
                                 application_id=auth_application_id)

        DiameterRequest._load(self, locals())


class SessionTerminationAnswer(DiameterAnswer):
    """Implementation of Session-Termination-Answer (STA) in Section 8.4.2 
    of IETF RFC 6733.

    The Session-Termination-Answer is indicated by the Command Code 275 and
    the Command Flags' 'R' bit cleared.

    Usage::

        >>> from bromelia.messages import SessionTerminationAnswer as STA
        >>> sta = STA()
        >>> sta
        <Diameter Message: 275 [STA], Default, 4 AVP(s)>
    """

    mandatory = {
                    "session_id": SessionIdAVP,
                    "result_code": ResultCodeAVP,
                    "origin_host": OriginHostAVP,
                    "origin_realm": OriginRealmAVP,
    }
    optionals = {
                    "user_name": UserNameAVP,
                    "_class": ClassAVP,
                    "error_message": ErrorMessageAVP,
                    "error_reporting_host": ErrorReportingHostAVP,
                    "failed_avp": FailedAvpAVP,
                    "origin_state_id": OriginStateIdAVP,
                    "redirect_host": RedirectHostAVP,
                    "redirect_host_usage": RedirectHostUsageAVP,
                    "redirect_max_cache_time": RedirectMaxCacheTimeAVP,
                    "proxy_info": ProxyInfoAVP
    }

    def __init__(self,
                session_id=platform.node(), 
                result_code=DIAMETER_SUCCESS, 
                origin_host=platform.node(), 
                origin_realm=socket.getfqdn(), 
                user_name=None,
                _class=None,
                error_message=None,
                error_reporting_host=None,
                failed_avp=None,
                origin_state_id=None,
                redirect_host=None,
                redirect_host_usage=None,
                redirect_max_cache_time=None,
                proxy_info=None,
                **kwargs):

        DiameterAnswer.__init__(self, 
                                command_code=SESSION_TERMINATION_MESSAGE,
                                application_id=DIAMETER_APPLICATION_DEFAULT)

        DiameterAnswer._load(self, locals())


class SessionTerminationRequest(DiameterRequest):
    """Implementation of Session-Termination-Request (STR) in Section 8.4.1 
    of IETF RFC 6733.

    The Session-Termination-Request is indicated by the Command Code 275 and
    the Command Flags' 'R' bit set.

    Usage::

        >>> from bromelia.messages import SessionTerminationRequest as STR
        >>> _str = STR()
        >>> _str
        <Diameter Message: 275 [STR], REQ Default, 6 AVP(s)>
    """

    mandatory = {
                    "session_id": SessionIdAVP,
                    "origin_host": OriginHostAVP,
                    "origin_realm": OriginRealmAVP,
                    "destination_realm": DestinationRealmAVP,
                    "auth_application_id": AuthApplicationIdAVP,
                    "termination_cause": TerminationCauseAVP,
    }
    optionals = {
                    "user_name": UserNameAVP,
                    "destination_host": DestinationHostAVP,
                    "_class": ClassAVP,
                    "origin_state_id": OriginStateIdAVP,
                    "proxy_info": ProxyInfoAVP,
                    "route_record": RouteRecordAVP
    }

    def __init__(self,
                session_id=platform.node(), 
                origin_host=platform.node(), 
                origin_realm=socket.getfqdn(), 
                destination_realm=socket.gethostbyname(platform.node()),
                auth_application_id=DIAMETER_APPLICATION_DEFAULT,
                termination_cause=DIAMETER_LOGOUT,
                user_name=None,
                destination_host=None,
                _class=None,
                origin_state_id=None,
                proxy_info=None,
                route_record=None,
                **kwargs):

        DiameterRequest.__init__(self, 
                                command_code=SESSION_TERMINATION_MESSAGE,
                                application_id=DIAMETER_APPLICATION_DEFAULT)

        DiameterRequest._load(self, locals())


# -*- coding: utf-8 -*-
"""
    bromelia.avps
    ~~~~~~~~~~~~~

    This module contains the Diameter protocol AVP library 
    that are used to create Diameter messages.
    
    :copyright: (c) 2020-present Henrique Marques Ribeiro.
    :license: MIT, see LICENSE for more details.
"""

from .etsi_3gpp.ts_129_061 import *
from .etsi_3gpp.ts_129_212 import *
from .etsi_3gpp.ts_129_214 import *
from .etsi_3gpp.ts_129_229 import *
from .etsi_3gpp.ts_129_272 import *
from .etsi_3gpp.ts_129_273 import *
from .etsi_3gpp.ts_129_329 import *
from .etsi_3gpp.ts_183_017 import *

from .ietf.rfc4006 import *
from .ietf.rfc4072 import *
from .ietf.rfc5447 import *
from .ietf.rfc6733 import *
from .ietf.rfc7155 import *
from .ietf.rfc8506 import *



# -*- coding: utf-8 -*-
"""
    bromelia.avps.ietf.rfc7155
    ~~~~~~~~~~~~~~~~~~~~~~~~~~

    This module contains Diameter AVP classes defined in IETF RFC 7155.
    
    :copyright: (c) 2020-present Henrique Marques Ribeiro.
    :license: MIT, see LICENSE for more details.
"""

from ...base import DiameterAVP
from ...constants.ietf.rfc7155 import *
from ...types import *


class FramedIpAddressAVP(DiameterAVP, AddressType):
    """Implementation of Framed-IP-Address AVP in Section 4.4.10.5.1 of 
    IETF RFC 7155.

    The Framed-IP-Address AVP (AVP Code 8) is of type Address.
    """
    code = FRAMED_IP_ADDRESS_AVP_CODE
    vendor_id = None

    def __init__(self, data):
        DiameterAVP.__init__(self, FramedIpAddressAVP.code)
        DiameterAVP.set_mandatory_bit(self, True)
        AddressType.__init__(self, data=data)


    def parser_data(self, data):
        if isinstance(data, str):
            ip_address = ipaddress.ip_address(data)

            if isinstance(ip_address, ipaddress.IPv4Address):
                self._data = ip_address.packed

            else:
                raise DataTypeError("AddressType MUST have data argument "\
                                    "of 'str' with IPv4 address format value")

        elif isinstance(data, bytes):
            self._data = data


class CalledStationIdAVP(DiameterAVP, UTF8StringType):
    """Implementation of Called-Station-Id AVP in Section 4.2.5 of
    IETF RFC 7155.

    The Called-Station-Id AVP (AVP Code 30) is of type UTF8String.
    """
    code = CALLED_STATION_ID_AVP_CODE
    vendor_id = None

    def __init__(self, data):
        DiameterAVP.__init__(self, CalledStationIdAVP.code)
        DiameterAVP.set_mandatory_bit(self, True)
        UTF8StringType.__init__(self, data=data)


class FramedIpv6PrefixAVP(DiameterAVP, OctetStringType):
    """Implementation of Framed-IPv6-Prefix AVP in Section 4.4.10.5.6 of 
    IETF RFC 7155.

    The Framed-IPv6-Prefix AVP (AVP Code 97) is of type OctetString.
    """
    code = FRAMED_IPV6_PREFIX_AVP_CODE
    vendor_id = None

    def __init__(self, data):
        DiameterAVP.__init__(self, FramedIpv6PrefixAVP.code)
        DiameterAVP.set_mandatory_bit(self, True)
        OctetStringType.__init__(self, data=data)


class CallingStationIdAVP(DiameterAVP, UTF8StringType):
    """Implementation of Calling-Station-Id AVP in Section 4.2.6 of 
    IETF RFC 7155.

    The Calling-Station-Id AVP (AVP Code 31) is of type UTF8String.
    """
    code = CALLING_STATION_ID_AVP_CODE
    vendor_id = None

    def __init__(self, data):
        DiameterAVP.__init__(self, CallingStationIdAVP.code)
        DiameterAVP.set_mandatory_bit(self, True)
        UTF8StringType.__init__(self, data=data)


# -*- coding: utf-8 -*-
"""
    bromelia.etsi_3gpp_swx.avps
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~

    This module contains the Diameter protocol AVP library that are used 
    to create Diameter messages for 3GPP SWs Application Id.
    
    :copyright: (c) 2020-present Henrique Marques Ribeiro.
    :license: MIT, see LICENSE for more details.
"""

from .._internal_utils import show_warn

show_warn("avps", "etsi_3gpp_swx")

from ..avps.etsi_3gpp.ts_129_212 import RatTypeAVP

from ..avps.etsi_3gpp.ts_129_272 import ContextIdentifierAVP
from ..avps.etsi_3gpp.ts_129_272 import ServiceSelectionAVP
from ..avps.etsi_3gpp.ts_129_272 import TerminalInformationAVP

from ..avps.etsi_3gpp.ts_129_273 import Non3gppUserDataAVP

from ..avps.etsi_3gpp.ts_129_229 import SipNumberAuthItemsAVP
from ..avps.etsi_3gpp.ts_129_229 import SipAuthenticationSchemeAVP
from ..avps.etsi_3gpp.ts_129_229 import SipAuthDataItemAVP
from ..avps.etsi_3gpp.ts_129_229 import ServerAssignmentTypeAVP
from ..avps.etsi_3gpp.ts_129_229 import SupportedFeaturesAVP
from ..avps.etsi_3gpp.ts_129_229 import ReasonCodeAVP
from ..avps.etsi_3gpp.ts_129_229 import DeregistrationReasonAVP
from ..avps.etsi_3gpp.ts_129_229 import VisitedNetworkIdentifierAVP

from ..avps.ietf.rfc5447 import Mip6AgentInfoAVP

from ..avps.ietf.rfc6733 import AuthApplicationIdAVP
from ..avps.ietf.rfc6733 import AuthSessionStateAVP
from ..avps.ietf.rfc6733 import DestinationHostAVP
from ..avps.ietf.rfc6733 import DestinationRealmAVP
from ..avps.ietf.rfc6733 import ExperimentalResultAVP
from ..avps.ietf.rfc6733 import OriginHostAVP
from ..avps.ietf.rfc6733 import OriginRealmAVP
from ..avps.ietf.rfc6733 import ResultCodeAVP
from ..avps.ietf.rfc6733 import SessionIdAVP
from ..avps.ietf.rfc6733 import UserNameAVP
from ..avps.ietf.rfc6733 import VendorIdAVP
from ..avps.ietf.rfc6733 import VendorSpecificApplicationIdAVP


#: -*- coding: utf-8 -*-
"""
    bromelia.constants.general
    ~~~~~~~~~~~~~~~~~~~~~~~~~~

    This module contains general constants defined in IETF RFC 6733.
    
    :copyright: (c) 2020-present Henrique Marques Ribeiro.
    :license: MIT, see LICENSE for more details.
"""

from .._internal_utils import convert_to_1_byte
from .._internal_utils import convert_to_4_bytes

#: General constants.
PRODUCT_NAME = f"Python bromelia"

DIAMETER_VERSION = convert_to_1_byte(1)
FIRMWARE_VERSION = convert_to_4_bytes(1)

#: Diameter Agent Types.
DIAMETER_AGENT_CLIENT_MODE = "CLIENT"
DIAMETER_AGENT_SERVER_MODE = "SERVER"

#: Basic AVP Data Formats.
AVP_LENGTH_UNSIGNED32 = 12
AVP_LENGTH_UNSIGNED32_AND_V_BIT = 16

AVP_LENGTH_INTEGER32 = 12
AVP_LENGTH_INTEGER32_AND_V_BIT = 16

AVP_LENGTH_INTEGER64 = 16
AVP_LENGTH_INTEGER64_AND_V_BIT = 20

#: Constants for Fields Lenght of Diameter Header.
DIAMETER_VERSION_FIELD = 1
DIAMETER_LENGTH_FIELD = 3
DIAMETER_FLAG_FIELD = 1
DIAMETER_COMMAND_CODE_FIELD = 3
DIAMETER_APPLICATION_ID_FIELD = 4
DIAMETER_HOP_BY_HOP_FIELD = 4
DIAMETER_END_TO_END_FIELD = 4

DIAMETER_HEADER_LENGTH = DIAMETER_VERSION_FIELD + \
                         DIAMETER_LENGTH_FIELD + \
                         DIAMETER_FLAG_FIELD + \
                         DIAMETER_COMMAND_CODE_FIELD + \
                         DIAMETER_APPLICATION_ID_FIELD + \
                         DIAMETER_HOP_BY_HOP_FIELD + \
                         DIAMETER_END_TO_END_FIELD

#: Constants for Fields Lenght of Diameter AVP Header.
AVP_CODE_FIELD = 4
AVP_FLAG_FIELD = 1
AVP_LENGTH_FIELD = 3
AVP_VENDOR_ID_FIELD = 4

AVP_HEADER_LENGTH = AVP_CODE_FIELD + AVP_FLAG_FIELD + AVP_LENGTH_FIELD
AVP_HEADER_LENGTH_LONGER = AVP_HEADER_LENGTH + AVP_VENDOR_ID_FIELD



# -*- coding: utf-8 -*-
"""
    bromelia.etsi_3gpp_gx.messages
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    This module contains the Diameter protocol messages for 3GPP Gx
    Application Id.

    :copyright: (c) 2020-present Henrique Marques Ribeiro.
    :license: MIT, see LICENSE for more details.
"""

import platform
import socket

from .avps import *

from .._internal_utils import show_warn
from ..base import DiameterRequest, DiameterAnswer
from ..constants import *

show_warn("messages", "etsi_3gpp_gx")


class CreditControlAnswer(DiameterAnswer):
    """Implementation of CC-Answer (CCA) command as per 
    clause 5.6.3 of ETSI TS 129 212 V15.3.0 (2018-07).

    The CC-Answer is indicated by the Command Code field 
    set to 272 and Command Flag's 'R' bit cleared.

    Usage::

        >>> from bromelia.etsi_3gpp_gx.messages import CreditControlAnswer as CCA
        >>> cca = CCA()
        >>> cca
        <Diameter Message: 272 [CCA] PXY, 16777238 [3GPP Gx], 7 AVP(s)>    
    """

    mandatory = {
                    "session_id": SessionIdAVP,
                    "auth_application_id": AuthApplicationIdAVP,
                    "origin_host": OriginHostAVP,
                    "origin_realm": OriginRealmAVP,
                    "cc_request_type": CcRequestTypeAVP,
                    "cc_request_number": CcRequestNumberAVP,
    }
    optionals = { 
                    # "drmp": DrmpAVP,
                    "result_code": ResultCodeAVP,
                    "experimental_result": ExperimentalResultAVP,
                    # "oc_supported_features": OcSupportedFeaturesAVP,
                    # "oc_olr": OcOlrAVP,
                    "supported_features": SupportedFeaturesAVP,
                    # "bearer_control_mode": BearerControlModeAVP,
                    "event_trigger": EventTriggerAVP,
                    # "event_report_indication": EventReportIndicationAVP,
                    "origin_state_id": OriginStateIdAVP,
                    "redirect_host": RedirectHostAVP,
                    "redirect_host_usage": RedirectHostUsageAVP,
                    "redirect_max_cache_time": RedirectMaxCacheTimeAVP,
                    # "charging_rule_remove": ChargingRuleRemoveAVP,
                    "charging_rule_install": ChargingRuleInstallAVP,
                    # "charging_information": ChargingInformationAVP,
                    "online": OnlineAVP,
                    "offline": OfflineAVP,
                    "qos_information": QosInformationAVP,
                    # "revalidation_time": RevalidationTimeAVP,
                    "default_eps_bearer_qos": DefaultEpsBearerQosAVP,
                    # "default_qos_information": DefaultQosInformationAVP,
                    "bearer_usage": BearerUsageAVP,
                    # "usage_monitoring_information": UsageMonitoringInformationAVP,
                    # "csg_information_reporting": CsgInformationReportingAVP,
                    # "user_csg_information": UserCsgInformationAVP,
                    # "pra_install": PraInstallAVP,
                    # "pra_remove": PraRemoveAVP,
                    # "presence_reporting_area_information": PresenceReportingAreaInformationAVP,
                    # "session_release_cause": SessionReleaseCauseAVP,
                    # "nbifom_support": NbifomSupportAVP,
                    # "nbifom_mode": NbifomModeAVP,
                    # "default_access": DefaultAccessAVP,
                    # "ran_rule_support": RanRuleSupportAVP,
                    # "routing_rule_report": RoutingRuleReportAVP,
                    # "conditional_policy_information": ConditionalPolicyInformationAVP,
                    # "removal_of_access": RemovalOfAccessAVP,
                    "ip_can_type": IpCanTypeAVP,
                    "error_message": ErrorMessageAVP,
                    # "error_reporting": ErrorReportingAVP,
                    "failed_avp": FailedAvpAVP,
                    "proxy_info": ProxyInfoAVP,
                    "route_record": RouteRecordAVP,
                    # "load": LoadAVP
    }

    def __init__(self,
                 session_id=platform.node(),
                 drmp=None,
                 auth_application_id=DIAMETER_APPLICATION_Gx,
                 origin_host=platform.node(),
                 origin_realm=socket.getfqdn(),
                 result_code=DIAMETER_SUCCESS,
                 experimental_result=None,
                 cc_request_type=CC_REQUEST_TYPE_INITIAL_REQUEST,
                 cc_request_number=0,
                 oc_supported_features=None,
                 oc_olr=None,
                 supported_features=None,
                 bearer_control_mode=None,
                 event_trigger=None,
                 event_report_indication=None,
                 origin_state_id=None,
                 redirect_host=None,
                 redirect_host_usage=None,
                 redirect_max_cache_time=None,
                 charging_rule_remove=None,
                 charging_rule_install=None,
                 charging_information=None,
                 online=None,
                 offline=None,
                 qos_information=None,
                 revalidation_time=None,
                 default_eps_bearer_qos=None,
                 default_qos_information=None,
                 bearer_usage=None,
                 usage_monitoring_information=None,
                 csg_information_reporting=None,
                 user_csg_information=None,
                 pra_install=None,
                 pra_remove=None,
                 presence_reporting_area_information=None,
                 session_release_cause=None,
                 nbifom_support=None,
                 nbifom_mode=None,
                 default_access=None,
                 ran_rule_support=None,
                 routing_rule_report=None,
                 conditional_policy_information=None,
                 removal_of_access=None,
                 ip_can_type=None,
                 error_message=None,
                 error_reporting_host=None,
                 failed_avp=None,
                 proxy_info=None,
                 route_record=None,
                 load=None,
                 **kwargs):

        DiameterAnswer.__init__(self,
                                 command_code=CC_MESSAGE,
                                 application_id=DIAMETER_APPLICATION_Gx)

        DiameterAnswer._load(self, locals())


class CreditControlRequest(DiameterRequest):
    """Implementation of CC-Request (CCR) command as per 
    clause 5.6.2 of ETSI TS 129 212 V15.3.0 (2018-07).

    The CC-Request is indicated by the Command Code field 
    set to 272 and the 'R' bit set in the Command Flags field.

    Usage::

        >>> from bromelia.etsi_3gpp_gx.messages import CreditControlRequest as CCR
        >>> from bromelia.constants import *
        >>> ccr_avps = {
        ...     "destination_realm": "remote",
        ...     "cc_request_type": CC_REQUEST_TYPE_INITIAL_REQUEST,
        ...     "cc_request_number": 1,
        ... }
        >>> ccr = CCR(**ccr_avps)
        >>> ccr
        <Diameter Message: 272 [CCR] REQ|PXY, 16777238 [3GPP Gx], 7 AVP(s)>
    """

    mandatory = {
                    "session_id": SessionIdAVP,
                    "auth_application_id": AuthApplicationIdAVP,
                    "origin_host": OriginHostAVP,
                    "origin_realm": OriginRealmAVP,
                    "destination_realm": DestinationRealmAVP,
                    "cc_request_type": CcRequestTypeAVP,
                    "cc_request_number": CcRequestNumberAVP,
    }
    optionals = {
                    # "drmp": DrmpAVP,
                    # "credit_management_status": CreditManagementStatusAVP,
                    "destination_host": DestinationHostAVP,
                    "origin_state_id": OriginStateIdAVP,
                    "subscription_id": SubscriptionIdAVP,
                    # "oc_supported_features": OcSupportedAVP,
                    "supported_features": SupportedFeaturesAVP,
                    # "tdf_information": TdfInformationAVP,
                    "network_request_support": NetworkRequestSupportAVP,
                    # "packet_filter_information": PacketFilterInformationAVP,
                    # "packet_filter_operation": PacketFilterOperationAVP,
                    # "bearer_identifier": BearerIdentifierAVP,
                    # "bearer_operation": BearerOperationAVP,
                    # "dynamic_address_flag": DynamicAddressFlagAVP,
                    # "dynamic_address_flag_extension": DynamicAddressFlagExtensionAVP,
                    # "pdn_connection_charging_id": PdnConnectionChargingIdAVP,
                    "framed_ip_address": FramedIpAddressAVP,
                    "framed_ipv6_prefix": FramedIpv6PrefixAVP,
                    "ip_can_type": IpCanTypeAVP,
                    # "x3gpp_rat_type": X3gppRatTypeAVP,
                    "an_trusted": AnTrustedAVP,
                    "rat_type": RatTypeAVP,
                    "termination_cause": TerminationCauseAVP,
                    "user_equipment_info": UserEquipmentInfoAVP,
                    "qos_information": QosInformationAVP,
                    # "qos_negotiation": QosNegotiationAVP,
                    # "qos_upgrade": QosUpgradeAVP,
                    "default_eps_bearer_qos": DefaultEpsBearerQosAVP,
                    # "default_qos_information": DefaultQosInformationAVP,
                    "an_gw_address": AnGwAddressAVP,
                    # "an_gw_status": AnGwStatusAVP,
                    # "x3gpp_sgsn_mcc_mnc": X3gppSgsnMccMncAVP,
                    # "x3gpp_sgsn_address": X3gppSgsnAddressAVP,
                    # "x3gpp_sgsn_ipv6_address": X3gppSgsnIpv6AddressAVP,
                    # "x3gpp_ggsn_address": X3gppGgsnAddressAVP,
                    # "x3gpp_ggsn_ipv6_address": X3gppGgsnIpv6AddressAVP,
                    # "x3gpp_selection_mode": X3gppSelectionModeAVP,
                    # "rai": RaiAVP,
                    # "x3gpp_user_location_info": X3gppUserLocationInfoAVP,
                    # "fixed_user_location_info": FixedUserLocationInfoAVP,
                    # "user_location_info_time": UserLocationInfoTimeAVP,
                    # "user_csg_information": UserCsgInformationAVP,
                    # "twan_identifier": TwanIdentifierAVP,
                    # "x3gpp_ms_timezone": X3gppMsTimezoneAVP,
                    # "ran_nas_release_cause": RanNasReleaseCauseAVP,
                    "x3gpp_charging_characteristics": X3gppChargingCharacteristicsAVP,
                    "called_station_id": CalledStationIdAVP,
                    # "pdn_connection_id": PdnConnectionIdAVP,
                    "bearer_usage": BearerUsageAVP,
                    "online": OnlineAVP,
                    "offline": OfflineAVP,
                    # "tft_packet_filter_information": TftPacketFilterInformationAVP,
                    "charging_rule_report": ChargingRuleReportAVP,
                    # "application_detection_information": ApplicationDetectionInformationAVP,
                    "event_trigger": EventTriggerAVP,
                    # "event_report_indication": EventReportIndicationAVP,
                    "access_network_charging_address": AccessNetworkChargingAddressAVP,
                    "access_network_charging_identifier_gx": AccessNetworkChargingIdentifierGxAVP,
                    # "coa_information": CoaInformationAVP,
                    # "usage_monitoring_information": UsageMonitoringInformationAVP,
                    # "nbifom_support": NbifomSupportAVP,
                    # "nbifom_mode": NbifomModeAVP,
                    # "default_access": DefaultAccessAVP,
                    # "origination_time_stamp": OriginationTimeStampAVP,
                    # "maximum_wait_time": MaximumWaitTimeAVP,
                    # "access_availability_change_reason": AccessAvailabilityChangeReasonAVP,
                    # "routing_rule_install": RoutingRuleInstallAVP,
                    # "routing_rule_remove": RoutingRuleRemoveAVP,
                    # "henb_local_ip_address": HenbLocalIpAddressAVP,
                    "ue_local_ip_address": UeLocalIpAddressAVP,
                    # "udp_source_port": UdpSourcePortAVP,
                    # "tcp_source_port": TcpLocalIpAddressAVP,
                    # "presence_reporting_area_information": PresenceReportingAreaInformationAVP,
                    # "logical_access_id": LogicalAccessIdAVP,
                    # "physical_access_id": PhysicalAccessIdAVP,
                    "proxy_info": ProxyInfoAVP,
                    "route_record": RouteRecordAVP,
                    # "x3gpp_ps_data_off_status": X3gppPsDataOffStatusAVP
    }

    def __init__(self,
                 session_id=platform.node(),
                 drmp=None,
                 auth_application_id=DIAMETER_APPLICATION_Gx,
                 origin_host=platform.node(),
                 origin_realm=socket.getfqdn(),
                 destination_realm=None,
                 cc_request_type=CC_REQUEST_TYPE_INITIAL_REQUEST,
                 cc_request_number=0,
                 credit_management_status=None,
                 destination_host=None,
                 origin_state_id=None,
                 subscription_id=None,
                 oc_supported_features=None,
                 supported_features=None,
                 tdf_information=None,
                 network_request_support=None,
                 packet_filter_information=None,
                 packet_filter_operation=None,
                 bearer_identifier=None,
                 bearer_operation=None,
                 dynamic_address_flag=None,
                 dynamic_address_flag_extension=None,
                 pdn_connection_charging_id=None,
                 framed_ip_address=None,
                 framed_ipv6_prefix=None,
                 ip_can_type=None,
                 x3gpp_rat_type=None,
                 an_trusted=None,
                 rat_type=None,
                 termination_cause=None,
                 user_equipment_info=None,
                 qos_information=None,
                 qos_negotiation=None,
                 qos_upgrade=None,
                 default_eps_bearer_qos=None,
                 default_qos_information=None,
                 an_gw_address=None,
                 an_gw_status=None,
                 x3gpp_sgsn_mcc_mnc=None,
                 x3gpp_sgsn_address=None,
                 x3gpp_sgsn_ipv6_address=None,
                 x3gpp_ggsn_address=None,
                 x3gpp_ggsn_ipv6_address=None,
                 x3gpp_selection_mode=None,
                 rai=None,
                 x3gpp_user_location_info=None,
                 fixed_user_location_info=None,
                 user_location_info_time=None,
                 user_csg_information=None,
                 twan_identifier=None,
                 x3gpp_ms_timezone=None,
                 ran_nas_release_cause=None,
                 x3gpp_charging_characteristics=None,
                 called_station_id=None,
                 pdn_connection_id=None,
                 bearer_usage=None,
                 online=None,
                 offline=None,
                 tft_packet_filter_information=None,
                 charging_rule_report=None,
                 application_detection_information=None,
                 event_trigger=None,
                 event_report_indication=None,
                 access_network_charging_address=None,
                 access_network_charging_identifier_gx=None,
                 coa_information=None,
                 usage_monitoring_information=None,
                 nbifom_support=None,
                 nbifom_mode=None,
                 default_access=None,
                 origination_time_stamp=None,
                 maximum_wait_time=None,
                 access_availability_change_reason=None,
                 routing_rule_install=None,
                 routing_rule_remove=None,
                 henb_local_ip_address=None,
                 ue_local_ip_address=None,
                 udp_source_port=None,
                 tcp_source_port=None,
                 presence_reporting_area_information=None,
                 logical_access_id=None,
                 physical_access_id=None,
                 proxy_info=None,
                 route_record=None,
                 x3gpp_ps_data_off_status=None,
                 **kwargs):

        DiameterRequest.__init__(self,
                                 command_code=CC_MESSAGE,
                                 application_id=DIAMETER_APPLICATION_Gx)

        DiameterRequest._load(self, locals())


class ReAuthAnswer(DiameterAnswer):
    """Implementation of Re-Auth-Answer (RAA) command as per 
    clause 5.6.5 of ETSI TS 129 212 V15.3.0 (2018-07).

    The Re-Auth-Answer is indicated by the Command Code field 
    set to 258 and Command Flag's 'R' bit cleared.

    Usage::

        >>> from bromelia.etsi_3gpp_gx.messages import ReAuthAnswer as RAA
        >>> raa = RAA()
        >>> raa
        <Diameter Message: 258 [RAA] PXY, 16777238 [3GPP Gx], 4 AVP(s)>  
    """

    mandatory = {
                    "session_id": SessionIdAVP,
                    "origin_host": OriginHostAVP,
                    "origin_realm": OriginRealmAVP,
    }
    optionals = { 
                    # "drmp": DrmpAVP,
                    "result_code": ResultCodeAVP,
                    "experimental_result": ExperimentalResultAVP,
                    "origin_state_id": OriginStateIdAVP,
                    # "oc_supported_features": OcSupportedFeaturesAVP,
                    # "oc_olr": OcOlrAVP,
                    "ip_can_type": IpCanTypeAVP,
                    "rat_type": RatTypeAVP,
                    "an_trusted": AnTrustedAVP,
                    "an_gw_address": AnGwAddressAVP,
                    # "x3gpp_sgsn_mcc_mnc": X3gppSgsnMccMncAVP,
                    # "x3gpp_sgsn_address": X3gppSgsnAddressAVP,
                    # "x3gpp_sgsn_ipv6_address": X3gppSgsnIpv6AddressAVP,
                    # "rai": RaiAVP,
                    # "x3gpp_user_location_info": X3gppUserLocationInfoAVP,
                    # "user_location_info_time": UserLocationInfoTimeAVP,
                    # "netloc_access_support": NetlocAccessSupportAVP,
                    # "user_csg_information": UserCsgInformationAVP,
                    # "x3gpp_ms_timezone": X3gppMsTimezoneAVP,
                    # "default_qos_information": DefaultQosInformationAVP,
                    "charging_rule_report": ChargingRuleReportAVP,
                    "error_message": ErrorMessageAVP,
                    "error_reporting_host": ErrorReportingHostAVP,
                    "failed_avp": FailedAvpAVP,
                    "proxy_info": ProxyInfoAVP
    }

    def __init__(self,
                 session_id=platform.node(),
                 drmp=None,
                 origin_host=platform.node(),
                 origin_realm=socket.getfqdn(),
                 result_code=DIAMETER_SUCCESS,
                 experimental_result=None,
                 origin_state_id=None,
                 oc_supported_features=None,
                 oc_olr=None,
                 ip_can_type=None,
                 rat_type=None,
                 an_trusted=None,
                 an_gw_address=None,
                 x3gpp_sgsn_mcc_mnc=None,
                 x3gpp_sgsn_address=None,
                 x3gpp_sgsn_ipv6_address=None,
                 rai=None,
                 x3gpp_user_location_info=None,
                 user_location_info_time=None,
                 netloc_access_support=None,
                 user_csf_information=None,
                 x3gpp_ms_timezone=None,
                 default_qos_information=None,
                 charging_rule_report=None,
                 error_message=None,
                 error_reporting_host=None,
                 failed_avp=None,
                 proxy_info=None,
                 **kwargs):

        DiameterAnswer.__init__(self,
                                 command_code=RE_AUTH_MESSAGE,
                                 application_id=DIAMETER_APPLICATION_Gx)

        DiameterAnswer._load(self, locals())


class ReAuthRequest(DiameterRequest):
    """Implementation of Re-Auth-Request (RAR) command as per 
    clause 5.6.4 of ETSI TS 129 212 V15.3.0 (2018-07).

    The Re-Auth-Request is indicated by the Command Code field 
    set to 258 and the 'R' bit set in the Command Flags field.

    Usage::

        >>> from bromelia.etsi_3gpp_gx.messages import ReAuthRequest as RAR
        >>> rar_avps = {
        ...     "destination_realm": "remote"
        ... }
        >>> rar = RAR(**rar_avps)
        >>> rar
        <Diameter Message: 258 [RAR] REQ|PXY, 16777238 [3GPP Gx], 6 AVP(s)>
    """

    mandatory = {
                    "session_id": SessionIdAVP,
                    "auth_application_id": AuthApplicationIdAVP,
                    "origin_host": OriginHostAVP,
                    "origin_realm": OriginRealmAVP,
                    "destination_realm": DestinationRealmAVP,
                    "re_auth_request_type": ReAuthRequestTypeAVP,
    }
    optionals = {
                    # "drmp": DrmpAVP,
                    # "session_release_cause": SessionReleaseCauseAVP,
                    "origin_state_id": OriginStateIdAVP,
                    # "oc_supported_features": OcSupportedFeaturesAVP,
                    "event_trigger": EventTriggerAVP,
                    # "event_report_indication": EventReportIndicationAVP,
                    # "charging_rule_remove": ChargingRuleRemoveAVP,
                    "charging_rule_install": ChargingRuleInstallAVP,
                    "default_eps_bearer_qos": DefaultEpsBearerQosAVP,
                    "qos_information": QosInformationAVP,
                    # "default_qos_information": DefaultQosInformationAVP,
                    # "revalidation_time": RevalidationTimeAVP,
                    # "usage_monitoring_information": UsageMonitoringInformationAVP,
                    # "pcscf_restoration_indication": PcscfRestorationIndicationAVP,
                    # "conditional_policy_information": ConditionalPolicyInformationAVP,
                    # "removal_of_access": RemovalOfAccessAVP,
                    "ip_can_type": IpCanTypeAVP,
                    # "pra_install": PraInstallAVP,
                    # "pra_remove": PraRemoveAVP,
                    # "csg_information_reporting": CsgInformationReportingAVP,
                    "proxy_info": ProxyInfoAVP,
                    "route_record": RouteRecordAVP
    }

    def __init__(self,
                 session_id=platform.node(),
                 drmp=None,
                 auth_application_id=DIAMETER_APPLICATION_Gx,
                 origin_host=platform.node(),
                 origin_realm=socket.getfqdn(),
                 destination_realm=None,
                 destination_host=None,
                 re_auth_request_type=RE_AUTH_REQUEST_TYPE_AUTHORIZE_ONLY,
                 session_release_cause=None,
                 origin_state_id=None,
                 oc_supported_features=None,
                 event_trigger=None,
                 event_report_indication=None,
                 charging_rule_remove=None,
                 charging_rule_install=None,
                 default_eps_bearer_qos=None,
                 qos_information=None,
                 default_qos_information=None,
                 revalidation_time=None,
                 usage_monitoring_information=None,
                 pcscf_restoration_indication=None,
                 conditional_policy_information=None,
                 removal_of_access=None,
                 ip_can_type=None,
                 pra_install=None,
                 pra_remove=None,
                 csg_information_reporting=None,
                 proxy_info=None,
                 route_record=None,
                 **kwargs):

        DiameterRequest.__init__(self,
                                 command_code=RE_AUTH_MESSAGE,
                                 application_id=DIAMETER_APPLICATION_Gx)

        DiameterRequest._load(self, locals())



# -*- coding: utf-8 -*-
"""
    bromelia.constants.experimental_result_codes
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    This module contains list of Diameter Experimental Result Codes defined
    in IETF RFC xxxx.
    
    :copyright: (c) 2020-present Henrique Marques Ribeiro.
    :license: MIT, see LICENSE for more details.
"""

DIAMETER_FIRST_REGISTRATION = bytes.fromhex("000007d1")                     #: 2001
DIAMETER_SUBSEQUENT_REGISTRATION = bytes.fromhex("000007d2")                #: 2002
DIAMETER_UNREGISTERED_SERVICE = bytes.fromhex("000007d3")                   #: 2003
DIAMETER_SUCCESS_SERVER_NAME_NOT_STORED = bytes.fromhex("000007d1")         #: 2004

DIAMETER_USER_DATA_NOT_AVAILABLE = bytes.fromhex("00001004")                #: 4100
DIAMETER_PRIOR_UPDATE_IN_PROGRESS = bytes.fromhex("00001005")               #: 4101
DIAMETER_AUTHENTICATION_DATA_UNAVAILABLE = bytes.fromhex("00001055")        #: 4181

DIAMETER_ERROR_USER_UNKNOWN = bytes.fromhex("00001389")                     #: 5001
DIAMETER_ERROR_IDENTITIES_DONT_MATCH = bytes.fromhex("0000138a")            #: 5002
DIAMETER_ERROR_IDENTITY_NOT_REGISTERED = bytes.fromhex("0000138b")          #: 5003
DIAMETER_ERROR_ROAMING_NOT_ALLOWED = bytes.fromhex("0000138c")              #: 5004
DIAMETER_ERROR_IDENTITY_ALREADY_REGISTERED = bytes.fromhex("0000138d")      #: 5005
DIAMETER_ERROR_AUTH_SCHEME_NOT_SUPPORTED = bytes.fromhex("0000138e")        #: 5006
DIAMETER_ERROR_IN_ASSIGNMENT_TYPE = bytes.fromhex("0000138f")               #: 5007
DIAMETER_ERROR_TOO_MUCH_DATA = bytes.fromhex("00001390")                    #: 5008
DIAMETER_ERROR_NOT_SUPPORTED_USER_DATA = bytes.fromhex("00001391")          #: 5009
DIAMETER_ERROR_FEATURE_UNSUPPORTED = bytes.fromhex("00001393")              #: 5011
DIAMETER_ERROR_SERVING_NODE_FEATURE_UNSUPPORTED = bytes.fromhex("00001394") #: 5012
DIAMETER_ERROR_USER_DATA_NOT_RECOGNIZED = bytes.fromhex("000013ec")         #: 5100
DIAMETER_ERROR_OPERATION_NOT_ALLOWED = bytes.fromhex("000013ed")            #: 5101
DIAMETER_ERROR_USER_DATA_CANNOT_BE_READ = bytes.fromhex("000013ee")         #: 5102
DIAMETER_ERROR_USER_DATA_CANNOT_BE_MODIFIED = bytes.fromhex("000013ef")     #: 5103
DIAMETER_ERROR_USER_DATA_CANNOT_BE_NOTIFIED = bytes.fromhex("000013f0")     #: 5104
DIAMETER_ERROR_TRANSPARENT_DATA_OUT_OF_SYNC = bytes.fromhex("000013f1")     #: 5105
DIAMETER_ERROR_SUBS_DATA_ABSENT = bytes.fromhex("000013f2")                 #: 5106
DIAMETER_ERROR_NO_SUBSCRIPTION_TO_DATA = bytes.fromhex("000013f3")          #: 5107
DIAMETER_ERROR_DSAI_NOT_AVAILABLE = bytes.fromhex("000013f4")               #: 5108
DIAMETER_ERROR_UNKNOWN_EPS_SUBSCRIPTION = bytes.fromhex("0000152c")         #: 5420
DIAMETER_ERROR_RAT_NOT_ALLOWED = bytes.fromhex("0000152d")                  #: 5421
DIAMETER_ERROR_EQUIPMENT_UNKNOWN = bytes.fromhex("0000152e")                #: 5422
DIAMETER_ERROR_UNKOWN_SERVING_NODE = bytes.fromhex("0000152f")              #: 5423
DIAMETER_ERROR_USER_NO_NON_3GPP_SUBSCRIPTION = bytes.fromhex("0000154a")    #: 5450
DIAMETER_ERROR_USER_NO_APN_SUBSCRIPTION = bytes.fromhex("0000154b")         #: 5451
DIAMETER_ERROR_RAT_TYPE_NOT_ALLOWED = bytes.fromhex("0000154c")             #: 5452



# -*- coding: utf-8 -*-
"""
    bromelia.lib.etsi_3gpp_s6a
    ~~~~~~~~~~~~~~~~~~~~~~~~~~

    A Python package with APIs for handling 3GPP S6a/S6d Application Id in the
    bromelia Python Library context.

    :copyright: (c) 2020-present Henrique Marques Ribeiro.
    :license: MIT, see LICENSE for more details.
"""

from .avps import *

from .messages import AuthenticationInformationAnswer as AIA
from .messages import AuthenticationInformationRequest as AIR

from .messages import CancelLocationAnswer as CLA
from .messages import CancelLocationRequest as CLR

from .messages import NotifyAnswer as NOA
from .messages import NotifyRequest as NOR

from .messages import PurgeUeAnswer as PUA
from .messages import PurgeUeRequest as PUR

from .messages import UpdateLocationAnswer as ULA
from .messages import UpdateLocationRequest as ULR


__title__ = "bromelia"
__description__ = "A Python micro framework for Diameter applications"
__url__ = "https://github.com/heimiricmr/bromelia"
__version__ = "0.3.1"
__author__ = "Henrique Marques Ribeiro"
__author_email__ = "heimiricmr@gmail.com"
__license__ = "MIT"
__copyright__ = "Copyright (c) 2020-present Henrique Marques Ribeiro"


# -*- coding: utf-8 -*-
"""
    bromelia.etsi_3gpp.ts_129_229
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    This module contains Diameter AVP classes defined in ETSI TS 129 229.
    
    :copyright: (c) 2020-present Henrique Marques Ribeiro.
    :license: MIT, see LICENSE for more details.
"""

from ..ietf.rfc6733 import VendorIdAVP

from ...base import DiameterAVP
from ...constants.etsi_3gpp.ts_129_229 import *
from ...types import *


class FeatureListIdAVP(DiameterAVP, Unsigned32Type):
    """Implementation of Feature-List-ID AVP in Section 6.3.30 of
    ETSI TS 129 229 V14.3.0 (2019-10).

    The Feature-List-ID AVP (AVP Code 629) is of type Unsigned32.
    """
    code = FEATURE_LIST_ID_AVP_CODE
    vendor_id = VENDOR_ID_3GPP

    def __init__(self, data):
        DiameterAVP.__init__(self, 
                             FeatureListIdAVP.code,
                             FeatureListIdAVP.vendor_id)
        DiameterAVP.set_vendor_id_bit(self, True)
        Unsigned32Type.__init__(self, data=data, vendor_id=VENDOR_ID_3GPP)


class FeatureListAVP(DiameterAVP, Unsigned32Type):
    """Implementation of Feature-List AVP in Section 6.3.31 of 
    ETSI TS 129 229 V14.3.0 (2019-10).

    The Feature-List AVP (AVP Code 630) is of type Unsigned32.
    """
    code = FEATURE_LIST_AVP_CODE
    vendor_id = VENDOR_ID_3GPP

    def __init__(self, data=convert_to_4_bytes(0)):
        DiameterAVP.__init__(self, 
                             FeatureListAVP.code,
                             FeatureListAVP.vendor_id)
        DiameterAVP.set_vendor_id_bit(self, True)
        Unsigned32Type.__init__(self, data=data, vendor_id=VENDOR_ID_3GPP)


class SupportedFeaturesAVP(DiameterAVP, GroupedType):
    """Implementation of Supported-Features AVP in Section 6.3.29 of 
    ETSI TS 129 229 V14.3.0 (2019-10).

    The Supported-Features AVP (AVP Code 628) is of type Grouped.
    """
    code = SUPPORTED_FEATURES_AVP_CODE
    vendor_id = VENDOR_ID_3GPP

    mandatory = {
                    "vendor_id": VendorIdAVP,
                    "feature_list_id": FeatureListIdAVP,
                    "feature_list": FeatureListAVP
    }
    optionals = {}

    def __init__(self, data):
        DiameterAVP.__init__(self, 
                             SupportedFeaturesAVP.code,
                             SupportedFeaturesAVP.vendor_id)
        DiameterAVP.set_vendor_id_bit(self, True)
        GroupedType.__init__(self, data=data, vendor_id=VENDOR_ID_3GPP)


class VisitedNetworkIdentifierAVP(DiameterAVP, OctetStringType):
    """Implementation of Visited-Network-Identifier AVP in Section 6.3.1  
    of ETSI TS 129 229 V15.2.0 (2019-10).

    The Visited-Network-Identifier AVP (AVP Code 600) is of type OctetString.
    """
    code = VISITED_NETWORK_IDENTIFIER_AVP_CODE
    vendor_id = VENDOR_ID_3GPP

    def __init__(self, data):
        DiameterAVP.__init__(self, 
                             VisitedNetworkIdentifierAVP.code,
                             VisitedNetworkIdentifierAVP.vendor_id)
        DiameterAVP.set_mandatory_bit(self, True)
        DiameterAVP.set_vendor_id_bit(self, True)
        OctetStringType.__init__(self, data=data, vendor_id=VENDOR_ID_3GPP)


class SipNumberAuthItemsAVP(DiameterAVP, Unsigned32Type):
    """Implementation of SIP-Number-Auth-Items AVP in Section 6.3.8 
    of ETSI TS 129 229 V16.2.0 (2020-11).

    The SIP-Number-Auth-Items AVP (AVP Code 607) is of type Unsigned32.
    """
    code = SIP_NUMBER_AUTH_ITEMS_AVP_CODE
    vendor_id = VENDOR_ID_3GPP

    def __init__(self, data):
        DiameterAVP.__init__(self, 
                             SipNumberAuthItemsAVP.code,
                             SipNumberAuthItemsAVP.vendor_id)
        DiameterAVP.set_mandatory_bit(self, True)
        DiameterAVP.set_vendor_id_bit(self, True)
        Unsigned32Type.__init__(self, data=data, vendor_id=VENDOR_ID_3GPP)


class SipAuthenticationSchemeAVP(DiameterAVP, UTF8StringType):
    """Implementation of Authentication-Scheme AVP in Section 6.3.9 
    of ETSI TS 129 229 V16.2.0 (2020-11).

    The Authentication-Scheme AVP (AVP Code 608) is of type UTF8String.
    """
    code = SIP_AUTHENTICATION_SCHEME_AVP_CODE
    vendor_id = VENDOR_ID_3GPP

    def __init__(self, data):
        DiameterAVP.__init__(self, 
                             SipAuthenticationSchemeAVP.code,
                             SipAuthenticationSchemeAVP.vendor_id)
        DiameterAVP.set_mandatory_bit(self, True)
        DiameterAVP.set_vendor_id_bit(self, True)
        UTF8StringType.__init__(self, data=data, vendor_id=VENDOR_ID_3GPP)


class SipAuthenticateAVP(DiameterAVP, OctetStringType):
    """Implementation of SIP-Authenticate AVP in Section 6.3.10 
    of ETSI TS 129 229 V16.2.0 (2020-11).

    The SIP-Authenticate AVP (AVP Code 609) is of type OctetString.
    """
    code = SIP_AUTHENTICATE_AVP_CODE
    vendor_id = VENDOR_ID_3GPP

    def __init__(self, data):
        DiameterAVP.__init__(self, 
                             SipAuthenticateAVP.code,
                             SipAuthenticateAVP.vendor_id)
        DiameterAVP.set_mandatory_bit(self, True)
        DiameterAVP.set_vendor_id_bit(self, True)
        OctetStringType.__init__(self, data=data, vendor_id=VENDOR_ID_3GPP)


class SipAuthorizationAVP(DiameterAVP, OctetStringType):
    """Implementation of SIP-Authorization AVP in Section 6.3.11 
    of ETSI TS 129 229 V16.2.0 (2020-11).

    The SIP-Authorization AVP (AVP Code 610) is of type OctetString.
    """
    code = SIP_AUTHORIZATION_AVP_CODE
    vendor_id = VENDOR_ID_3GPP

    def __init__(self, data):
        DiameterAVP.__init__(self, 
                             SipAuthorizationAVP.code,
                             SipAuthorizationAVP.vendor_id)
        DiameterAVP.set_mandatory_bit(self, True)
        DiameterAVP.set_vendor_id_bit(self, True)
        OctetStringType.__init__(self, data=data, vendor_id=VENDOR_ID_3GPP)


class ConfidentialityKeyAVP(DiameterAVP, OctetStringType):
    """Implementation of Confidentiality-Key AVP in Section 6.3.27 
    of ETSI TS 129 229 V16.2.0 (2020-11).

    The Confidentiality-Key AVP (AVP Code 625) is of type OctetString.
    """
    code = CONFIDENTIALITY_KEY_AVP_CODE
    vendor_id = VENDOR_ID_3GPP

    def __init__(self, data):
        DiameterAVP.__init__(self, 
                             ConfidentialityKeyAVP.code,
                             ConfidentialityKeyAVP.vendor_id)
        DiameterAVP.set_mandatory_bit(self, True)
        DiameterAVP.set_vendor_id_bit(self, True)
        OctetStringType.__init__(self, data=data, vendor_id=VENDOR_ID_3GPP)


class IntegrityKeyAVP(DiameterAVP, OctetStringType):
    """Implementation of Integrity-Key AVP in Section 6.3.28
    of ETSI TS 129 229 V16.2.0 (2020-11).

    The Integrity-Key AVP (AVP Code 626) is of type OctetString.
    """
    code = INTEGRITY_KEY_AVP_CODE
    vendor_id = VENDOR_ID_3GPP

    def __init__(self, data):
        DiameterAVP.__init__(self, 
                             IntegrityKeyAVP.code,
                             IntegrityKeyAVP.vendor_id)
        DiameterAVP.set_mandatory_bit(self, True)
        DiameterAVP.set_vendor_id_bit(self, True)
        OctetStringType.__init__(self, data=data, vendor_id=VENDOR_ID_3GPP)


class SipAuthDataItemAVP(DiameterAVP, GroupedType):
    """Implementation of SIP-Auth-Data-Item AVP in Section 6.3.13 
    of ETSI TS 129 229 V16.2.0 (2020-11).

    The SIP-Auth-Data-Item AVP (AVP Code 612) is of type Grouped.
    """
    code = SIP_AUTH_DATA_ITEM_AVP_CODE
    vendor_id = VENDOR_ID_3GPP

    optionals = {
                    #"sip_item_number": SipItemNumberAVP,
                    "sip_authentication_scheme": SipAuthenticationSchemeAVP,
                    "sip_authenticate": SipAuthenticateAVP,
                    "sip_authorization": SipAuthorizationAVP,
                    #"sip_authentication_context": SipAuthenticationContextAVP,
                    "confidentiality_key": ConfidentialityKeyAVP,
                    "integrity_key": IntegrityKeyAVP,
                    #"sip_digest_authenticate": SipDigestAuthenticateAVP,
                    #"framed_ip_address": FramedIpvAddressAVP,
                    #"framed_ipv6_address": FramedIpv6AddressAVP,
                    #"framed_interface_id": FramedInterfaceIdAVP,
                    #"line_identifier": LineIdentifierAVP,
    }

    def __init__(self, data):
        DiameterAVP.__init__(self, 
                             SipAuthDataItemAVP.code,
                             SipAuthDataItemAVP.vendor_id)
        DiameterAVP.set_mandatory_bit(self, True)
        DiameterAVP.set_vendor_id_bit(self, True)
        GroupedType.__init__(self, data=data, vendor_id=VENDOR_ID_3GPP)


class ServerAssignmentTypeAVP(DiameterAVP, EnumeratedType):
    """Implementation of Server-Assignment-Type AVP in Section 6.3.15 
    of ETSI TS 129 229 V16.2.0 (2020-11).

    The Server-Assignment-Type AVP (AVP Code 614) is of type Enumerated.
    """
    code = SERVER_ASSIGNMENT_TYPE_AVP_CODE
    vendor_id = VENDOR_ID_3GPP

    values = [
                SERVER_ASSIGNMENT_TYPE_NO_ASSIGNMENT,
                SERVER_ASSIGNMENT_TYPE_REGISTRATION,
                SERVER_ASSIGNMENT_TYPE_RE_REGISTRATION,
                SERVER_ASSIGNMENT_TYPE_UNREGISTERED_USER,
                SERVER_ASSIGNMENT_TYPE_TIMEOUT_DEREGISTRATION,
                SERVER_ASSIGNMENT_TYPE_USER_DEREGISTRATION,
                SERVER_ASSIGNMENT_TYPE_DEREGISTRATION_STORE_SERVER_NAME,
                SERVER_ASSIGNMENT_TYPE_USER_DEREGISTRATION_STORE_SERVER_NAME,
                SERVER_ASSIGNMENT_TYPE_ADMINISTRATIVE_DEREGISTRATION,
                SERVER_ASSIGNMENT_TYPE_AUTHENTICATION_FAILURE,
                SERVER_ASSIGNMENT_TYPE_AUTHENTICATION_TIMEOUT,
                SERVER_ASSIGNMENT_TYPE_DEREGISTRATION_TOO_MUCH_DATA,
                SERVER_ASSIGNMENT_TYPE_AAA_USER_DATA_REQUEST,
                SERVER_ASSIGNMENT_TYPE_PGW_UPDATE,
                SERVER_ASSIGNMENT_TYPE_RESTORATION
    ]

    def __init__(self, data):
        DiameterAVP.__init__(self, 
                             ServerAssignmentTypeAVP.code,
                             ServerAssignmentTypeAVP.vendor_id)
        DiameterAVP.set_mandatory_bit(self, True)
        DiameterAVP.set_vendor_id_bit(self, True)
        EnumeratedType.__init__(self, data=data, vendor_id=VENDOR_ID_3GPP)


class ReasonCodeAVP(DiameterAVP, EnumeratedType):
    """Implementation of Reason-Code AVP in Section 6.3.17 
    of ETSI TS 129 229 V11.3.0 (2013-04).

    The Reason-Code AVP (AVP Code 616) is of type Enumerated.
    """
    code = REASON_CODE_AVP_CODE
    vendor_id = VENDOR_ID_3GPP

    values = [
                REASON_CODE_PERMANENT_TERMINATION,
                REASON_CODE_NEW_SERVER_ASSIGNED,
                REASON_CODE_SERVER_CHANGE,
                REASON_CODE_REMOVE_CHANGE,
    ]

    def __init__(self, data):
        DiameterAVP.__init__(self, 
                             ReasonCodeAVP.code,
                             ReasonCodeAVP.vendor_id)
        DiameterAVP.set_mandatory_bit(self, True)
        DiameterAVP.set_vendor_id_bit(self, True)
        EnumeratedType.__init__(self, data=data, vendor_id=VENDOR_ID_3GPP)


class ReasonInfoAVP(DiameterAVP, UTF8StringType):
    """Implementation of Reason-Info AVP in Section 6.3.18
    of ETSI TS 129 229 V11.3.0 (2013-04).

    The Reason-Info AVP (AVP Code 617) is of type UTF8String.
    """
    code = REASON_INFO_AVP_CODE
    vendor_id = VENDOR_ID_3GPP

    def __init__(self, data):
        DiameterAVP.__init__(self, 
                             ReasonInfoAVP.code,
                             ReasonInfoAVP.vendor_id)
        DiameterAVP.set_mandatory_bit(self, True)
        DiameterAVP.set_vendor_id_bit(self, True)
        UTF8StringType.__init__(self, data=data, vendor_id=VENDOR_ID_3GPP)


class DeregistrationReasonAVP(DiameterAVP, GroupedType):
    """Implementation of Deregistration-Reason AVP in Section 6.3.16 
    of ETSI TS 129 229 V11.3.0 (2013-04).

    The Server-Assignment-Type AVP (AVP Code 615) is of type Grouped.
    """
    code = DE_REGISTRATION_REASON_AVP_CODE
    vendor_id = VENDOR_ID_3GPP

    mandatory = {
                    "reason_code": ReasonCodeAVP,
    }
    optionals = {
                    "reason_info": ReasonInfoAVP,
    }

    def __init__(self, data):
        DiameterAVP.__init__(self, 
                             DeregistrationReasonAVP.code,
                             DeregistrationReasonAVP.vendor_id)
        DiameterAVP.set_mandatory_bit(self, True)
        DiameterAVP.set_vendor_id_bit(self, True)
        GroupedType.__init__(self, data=data, vendor_id=VENDOR_ID_3GPP)


# -*- coding: utf-8 -*-
"""
    bromelia.messages
    ~~~~~~~~~~~~~~~~~

    This module contains the Diameter protocol base messages.

    :copyright: (c) 2020-present Henrique Marques Ribeiro.
    :license: MIT, see LICENSE for more details.
"""

from .lib.ietf_rfc6733.messages import AbortSessionAnswer as ASA
from .lib.ietf_rfc6733.messages import AbortSessionRequest as ASR

from .lib.ietf_rfc6733.messages import CapabilitiesExchangeAnswer as CEA
from .lib.ietf_rfc6733.messages import CapabilitiesExchangeRequest as CER

from .lib.ietf_rfc6733.messages import DeviceWatchdogAnswer as DWA
from .lib.ietf_rfc6733.messages import DeviceWatchdogRequest as DWR

from .lib.ietf_rfc6733.messages import DisconnectPeerAnswer as DPA
from .lib.ietf_rfc6733.messages import DisconnectPeerRequest as DPR

from .lib.ietf_rfc6733.messages import ReAuthAnswer as RAA
from .lib.ietf_rfc6733.messages import ReAuthRequest as RAR

from .lib.ietf_rfc6733.messages import SessionTerminationAnswer as STA
from .lib.ietf_rfc6733.messages import SessionTerminationRequest as STR

from .lib.ietf_rfc6733.messages import *


# -*- coding: utf-8 -*-
"""
    bromelia.lib.etsi_3gpp_swx
    ~~~~~~~~~~~~~~~~~~~~~~~~~~

    A Python package with APIs for handling 3GPP SWx Application Id in the
    bromelia Python Library context.

    :copyright: (c) 2020-present Henrique Marques Ribeiro.
    :license: MIT, see LICENSE for more details.
"""

from .avps import *

from .messages import MultimediaAuthAnswer as MAA
from .messages import MultimediaAuthRequest as MAR

from .messages import RegistrationTerminationAnswer as RTA
from .messages import RegistrationTerminationRequest as RTR

from .messages import ServerAssignmentAnswer as SAA
from .messages import ServerAssignmentRequest as SAR


# -*- coding: utf-8 -*-
"""
    bromelia.constants.ietf.rfc6733
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    This module contains constants defined in IETF RFC 6733.
 
    :copyright: (c) 2020-present Henrique Marques Ribeiro.
    :license: MIT, see LICENSE for more details.
"""

from ..._internal_utils import convert_to_2_bytes
from ..._internal_utils import convert_to_3_bytes
from ..._internal_utils import convert_to_4_bytes


#: Diameter Messages
CAPABILITIES_EXCHANGE_MESSAGE = convert_to_3_bytes(257)
ABORT_SESSION_MESSAGE = convert_to_3_bytes(274)
DEVICE_WATCHDOG_MESSAGE = convert_to_3_bytes(280)
DISCONNECT_PEER_MESSAGE = convert_to_3_bytes(282)

#: Diameter AVPs
USER_NAME_AVP_CODE = convert_to_4_bytes(1)
CLASS_AVP_CODE = convert_to_4_bytes(25)
SESSION_TIMEOUT_AVP_CODE = convert_to_4_bytes(27)
PROXY_STATE_AVP_CODE = convert_to_4_bytes(33)
ACCT_SESSION_ID_AVP_CODE = convert_to_4_bytes(44)
ACCT_MULTI_SESSION_ID_AVP_CODE = convert_to_4_bytes(50)
EVENT_TIMESTAMP_AVP_CODE = convert_to_4_bytes(55)
ACCT_INTERIM_INTERVAL_AVP_CODE = convert_to_4_bytes(85)
HOST_IP_ADDRESS_AVP_CODE = convert_to_4_bytes(257)
AUTH_APPLICATION_ID_AVP_CODE = convert_to_4_bytes(258)
ACCT_APPLICATION_ID_AVP_CODE = convert_to_4_bytes(259)
VENDOR_SPECIFIC_APPLICATION_ID_AVP_CODE = convert_to_4_bytes(260)
REDIRECT_HOST_USAGE_AVP_CODE = convert_to_4_bytes(261)
REDIRECT_MAX_CACHE_TIME_AVP_CODE = convert_to_4_bytes(262)
SESSION_ID_AVP_CODE = convert_to_4_bytes(263)
ORIGIN_HOST_AVP_CODE = convert_to_4_bytes(264)
SUPPORTED_VENDOR_ID_AVP_CODE = convert_to_4_bytes(265)
VENDOR_ID_AVP_CODE = convert_to_4_bytes(266)
FIRMWARE_REVISION_AVP_CODE = convert_to_4_bytes(267)
RESULT_CODE_AVP_CODE = convert_to_4_bytes(268)
PRODUCT_NAME_AVP_CODE = convert_to_4_bytes(269)
SESSION_BINDING_AVP_CODE = convert_to_4_bytes(270)
SESSION_SERVER_FAILOVER_AVP_CODE = convert_to_4_bytes(271)
MULTI_ROUND_TIME_OUT_AVP_CODE = convert_to_4_bytes(272)
DISCONNECT_CAUSE_AVP_CODE = convert_to_4_bytes(273)
AUTH_REQUEST_TYPE_AVP_CODE = convert_to_4_bytes(274)
AUTH_GRACE_PERIOD_AVP_CODE = convert_to_4_bytes(276)
AUTH_SESSION_STATE_AVP_CODE = convert_to_4_bytes(277)
ORIGIN_STATE_ID_AVP_CODE = convert_to_4_bytes(278)
FAILED_AVP_AVP_CODE = convert_to_4_bytes(279)
PROXY_HOST_AVP_CODE = convert_to_4_bytes(280)
ERROR_MESSAGE_AVP_CODE = convert_to_4_bytes(281)
ROUTE_RECORD_AVP_CODE = convert_to_4_bytes(282)
DESTINATION_REALM_AVP_CODE = convert_to_4_bytes(283)
PROXY_INFO_AVP_CODE = convert_to_4_bytes(284)
RE_AUTH_REQUEST_TYPE_AVP_CODE = convert_to_4_bytes(285)
ACCOUNTING_SUB_SESSION_ID_AVP_CODE = convert_to_4_bytes(287)
AUTHORIZATION_LIFETIME_AVP_CODE = convert_to_4_bytes(291)
REDIRECT_HOST_AVP_CODE = convert_to_4_bytes(292)
DESTINATION_HOST_AVP_CODE = convert_to_4_bytes(293)
ERROR_REPORTING_HOST_AVP_CODE = convert_to_4_bytes(294)
TERMINATION_CAUSE_AVP_CODE = convert_to_4_bytes(295)
ORIGIN_REALM_AVP_CODE = convert_to_4_bytes(296)
EXPERIMENTAL_RESULT_AVP_CODE = convert_to_4_bytes(297)
EXPERIMENTAL_RESULT_CODE_AVP_CODE = convert_to_4_bytes(298)
INBAND_SECURITY_ID_AVP_CODE = convert_to_4_bytes(299)
ACCOUNTING_RECORD_TYPE_AVP_CODE = convert_to_4_bytes(480)
ACCOUNTING_REALTIME_REQUIRED_AVP_CODE = convert_to_4_bytes(483)
ACCOUNTING_RECORD_NUMBER_AVP_CODE = convert_to_4_bytes(485)

#: List of Host-IP-Address types.
#: For more information, please refer to Section 5.3.5 of IETF RFC 6733.
HOST_IP_ADDRESS_FAMILY_CODE_IPV4 = convert_to_2_bytes(1)
HOST_IP_ADDRESS_FAMILY_CODE_IPV6 = convert_to_2_bytes(2)

#: List of Vendor-Id values.
#: For more information, please refer to Section 5.3.3 of IETF RFC 6733.
VENDOR_ID_DEFAULT = convert_to_4_bytes(0)
VENDOR_ID_TEKELEC = convert_to_4_bytes(323)
VENDOR_ID_3GPP = convert_to_4_bytes(10415)
VENDOR_ID_ETSI = convert_to_4_bytes(13019)

#: List of Disconnect-Cause AVP values.
#: For more information, please refer to Section 5.4.3 of IETF RFC 6733.
DISCONNECT_CAUSE_REBOOTING = convert_to_4_bytes(0)
DISCONNECT_CAUSE_BUSY = convert_to_4_bytes(1)
DISCONNECT_CAUSE_DO_NOT_WANT_TO_TALK_TO_YOU = convert_to_4_bytes(2)

#: List of Inband-Security-Id AVP values.
#: For more information, please refer to Section 6.10 of IETF RFC 6733.
INBAND_SECURITY_ID_NO_SECURITY = convert_to_4_bytes(0)
INBAND_SECURITY_ID_TLS = convert_to_4_bytes(1)

#: List of Redirect-Host-Usage AVP values.
#: For more information, please refer to Section 6.13 of IETF RFC 6733.
REDIRECT_HOST_USAGE_DONT_CACHE = convert_to_4_bytes(0)
REDIRECT_HOST_USAGE_ALL_SESSION = convert_to_4_bytes(1)
REDIRECT_HOST_USAGE_ALL_REALM = convert_to_4_bytes(2)
REDIRECT_HOST_USAGE_REALM_AND_APPLICATION = convert_to_4_bytes(3)
REDIRECT_HOST_USAGE_ALL_APPLICATION = convert_to_4_bytes(4)
REDIRECT_HOST_USAGE_ALL_HOST = convert_to_4_bytes(5)
REDIRECT_HOST_USAGE_ALL_USER = convert_to_4_bytes(6)

#: List of Auth-Request-Type AVP values.
#: For more information, please refer to Section 8.7 of IETF RFC 6733.
AUTH_REQUEST_TYPE_AUTHENTICATE_ONLY = convert_to_4_bytes(1)
AUTH_REQUEST_TYPE_AUTHORIZE_ONLY = convert_to_4_bytes(2)
AUTH_REQUEST_TYPE_AUTHORIZE_AUTHENTICATE = convert_to_4_bytes(3)

#: List of Auth-Session-State AVP values.
#: For more information, please refer to Section 8.11 of IETF RFC 6733.
STATE_MAINTAINED = convert_to_4_bytes(0)
NO_STATE_MAINTAINED = convert_to_4_bytes(1)

#: List of Re-Auth-Request-Type AVP values.
#: For more information, please refer to Section 8.12 of IETF RFC 6733.
RE_AUTH_REQUEST_TYPE_AUTHORIZE_ONLY = convert_to_4_bytes(0)
RE_AUTH_REQUEST_TYPE_AUTHORIZE_AUTHENTICATE = convert_to_4_bytes(1)

#: List of Session-Server-Failover AVP values.
#: For more information, please refer to Section 8.18 of IETF RFC 6733.
SESSION_SERVER_FAILOVER_REFUSE_SERVICE = convert_to_4_bytes(0)
SESSION_SERVER_FAILOVER_TRY_AGAIN = convert_to_4_bytes(1)
SESSION_SERVER_FAILOVER_ALLOW_SERVICE = convert_to_4_bytes(2)
SESSION_SERVER_FAILOVER_TRY_AGAIN_ALLOW_SERVICE = convert_to_4_bytes(3)

#: List of Termination-Cause AVP values.
#: For more information, please refer to IANA's website or Section 8.47 
#: of IETF RFC 6733.
#: https://www.iana.org/assignments/aaa-parameters/aaa-parameters.xml#aaa-parameters-16
DIAMETER_LOGOUT = convert_to_4_bytes(1)
DIAMETER_SERVICE_NOT_PROVIDED = convert_to_4_bytes(2)
DIAMETER_BAD_ANSWER = convert_to_4_bytes(3)
DIAMETER_ADMINISTRATIVE = convert_to_4_bytes(4)
DIAMETER_LINK_BROKEN = convert_to_4_bytes(5)
DIAMETER_AUTH_EXPIRED = convert_to_4_bytes(6)
DIAMETER_USER_MOVED = convert_to_4_bytes(7)
DIAMETER_SESSION_TIMEOUT = convert_to_4_bytes(8)

#: List of Accounting-Record-Type AVP values.
#: For more information, please refer to Section 9.8.1 of IETF RFC 6733.
ACCOUNTING_RECORD_TYPE_EVENT_RECORD = convert_to_4_bytes(1)
ACCOUNTING_RECORD_TYPE_START_RECORD = convert_to_4_bytes(2)
ACCOUNTING_RECORD_TYPE_INTERIM_RECORD = convert_to_4_bytes(3)
ACCOUNTING_RECORD_TYPE_STOP_RECORD = convert_to_4_bytes(4)

#: List of Accounting-Realtime-Required AVP values.
#: For more information, please refer to Section 9.8.7 of IETF RFC 6733.
ACCOUNTING_REALTIME_REQUIRED_DELIVER_AND_GRANT = convert_to_4_bytes(1)
ACCOUNTING_REALTIME_REQUIRED_GRANT_AND_STORE = convert_to_4_bytes(2)
ACCOUNTING_REALTIME_REQUIRED_GRAND_AND_LOSE = convert_to_4_bytes(3)



# -*- coding: utf-8 -*-
"""
    bromelia.bromelia
    ~~~~~~~~~~~~~~~~~

    This module contains the orchestrator to allow a seamless
    Diameter application implementation.

    :copyright: (c) 2020-present Henrique Marques Ribeiro.
    :license: MIT, see LICENSE for more details.
"""

import logging
import multiprocessing
import sys
import threading
import time
from copy import copy
from types import SimpleNamespace

from ._internal_utils import _convert_file_to_config
from ._internal_utils import application_id_look_up
from ._internal_utils import get_app_name
from .avps import DestinationHostAVP
from .avps import DestinationRealmAVP
from .avps import OriginHostAVP
from .avps import OriginRealmAVP
from .avps import ResultCodeAVP
from .avps import SessionIdAVP
from .base import DiameterAnswer
from .base import DiameterRequest
from .config import *
from .constants import *
from .exceptions import BromeliaException
from .setup import Diameter
from .utils import is_3xxx_failure
from .utils import is_4xxx_failure
from .utils import is_5xxx_failure


worker_logger = logging.getLogger("Worker")
bromelia_logger = logging.getLogger("Bromelia")


def get_application_string_by_id(application_id):
    applications = [ application for application in globals().items() if "DIAMETER_APPLICATION_" in application[0] ]
    for _string, _id in applications:
        if _id == application_id:
            return _string.split("DIAMETER_APPLICATION_")[1].lower()


def get_formatted_answer_as_per_request(answer, request):
    answer.header.application_id = request.header.application_id
    answer.header.hop_by_hop = request.header.hop_by_hop
    answer.header.end_to_end = request.header.end_to_end

    if request.has_avp("session_id_avp"):
        answer.session_id_avp.data = request.session_id_avp.data
        answer.refresh()
    return answer


class Global(object):
    """Helper class to allow the creation of Bromelia attributes on the go. It
    works as a namespace.
    """
    pass


class PendingAnswer:
    """Helper class to allow the pending answer tracking of expected Diameter
    Answers for a set of Diameter Requests. It supports the Bromelia class inner 
    workings.
    """
    def __init__(self):
        self.recv_event = threading.Event()
        self.stop_event = threading.Event()
        self.message = None


class Worker(multiprocessing.Process):
    associations = dict()
    recv_queues = list()


    def __init__(self, app, manager):
        multiprocessing.Process.__init__(self)
        self.daemon = True

        self.name = Worker.set_name(app.config["APPLICATIONS"])
        self.app = app
        self.is_open = manager.Event()

        self.recv_queue = manager.Queue()
        self.send_queue = manager.Queue()

        self.send_event = manager.Event()
        self.send_lock = manager.Lock()
        self.recv_lock = manager.Lock()

        self.pending_answers = dict()

        self.update_associations()
        self.update_recv_queues()

        worker_logger.debug(f"Initializing Worker {self}")


    @staticmethod
    def set_name(applications):
        name = ""
        for application in applications:
            app_name = application_id_look_up(application["app_id"])[0]
            name += f'{app_name};'
        return name[:-1].replace(" ", "_")


    def update_associations(self):
        for application in self.app.config["APPLICATIONS"]:
            association = {application["app_id"]: self}
            Worker.associations.update(association)


    def update_recv_queues(self):
        Worker.recv_queues.append([self.recv_queue, self.recv_lock])


    def send_handler(self):
        while True:
            if self.send_queue.empty():
                worker_logger.debug("The send_queue is empty and there is no "\
                                    "notification to go ahead")

                self.send_event.wait(timeout=1)

                worker_logger.debug("Got go ahead for send_event Event")


            worker_logger.debug(f"[{self.name}] There is/are "\
                                f"{self.send_queue.qsize()} message(s) "\
                                f"available to be sent")

            if self.send_queue.qsize() == 1:
                outgoing_message = self.send_queue.get()
                hop_by_hop = outgoing_message.header.hop_by_hop

                worker_logger.debug(f"[{self.name}][{hop_by_hop.hex()}] "\
                                    f"Sending to endpoint")

                self.app.send_message(outgoing_message)

                self.send_event.clear()
                self.send_lock.release()

            elif self.send_queue.qsize() > 1:
                outgoing_messages = list()

                while not self.send_queue.empty():
                    outgoing_message = self.send_queue.get()
                    hop_by_hop = outgoing_message.header.hop_by_hop

                    outgoing_messages.append(outgoing_message)
                
                    worker_logger.debug(f"[{self.name}][{hop_by_hop.hex()}] "\
                                        f"Sending to endpoint")

                self.app.send_messages(outgoing_messages)

                self.send_event.clear()
                self.send_lock.release()


    def recv_handler(self):
        while True:
            incoming_message = self.app.get_message()

            if incoming_message:
                user_name = None
                if incoming_message.has_avp("user_name_avp"):
                    user_name = incoming_message.user_name_avp.data

                worker_logger.debug(f"Message from Diameter Layer Process: "\
                                    f"{application_id_look_up(incoming_message.header.application_id)[0]}, "\
                                    f"{incoming_message.header.get_command_code()}, "\
                                    f"{incoming_message.header.hop_by_hop.hex()}, "\
                                    f"DATA LENGTH: {len(incoming_message)}, "\
                                    f"USERNAME: {user_name}")
    
                hop_by_hop = incoming_message.header.hop_by_hop

                self.recv_lock.acquire()
                self.recv_queue.put(incoming_message)
                self.recv_lock.release()
    
                worker_logger.debug(f"[{self.name}][{hop_by_hop.hex()}] "\
                                    f"Putting into recv_queue Queue")


    #: it starts under worker.start() call
    def run(self):
        with self.app.context():
            try:
                while self.app.is_open():
                    self.is_open.set()

                    recv_thrd = threading.Thread(name="recv_handler", 
                                                 target=self.recv_handler, 
                                                 daemon=True)

                    send_thrd = threading.Thread(name="send_handler", 
                                                 target=self.send_handler, 
                                                 daemon=True)

                    recv_thrd.start()
                    send_thrd.start()

                    recv_thrd.join()
                    send_thrd.join()

            except KeyboardInterrupt:
                sys.exit(0)

            self.is_open.clear()


    def is_running(self):
        return self.is_open.is_set()


class Bromelia:
    """The Bromelia object implements a WSGI-like application but for Diameter 
    protocol and acts as the central object. It will spin up one or more 
    processes by using the Worker class. The number of processes depends on the 
    number of Diameter interfaces needed for a given Diameter application. 
    
    Each process created by running Worker objects will represent a single 
    Diameter interface sending to / receiving byte stream from a Peer Node. Such
    traffic will be proxied to the centralized process which holds the Bromelia 
    object. The Bromelia object is responsible for process either Diameter 
    requests or Diameter answers according to the Diameter interface and 
    properly forward it to the expected Diameter interface.

    It is strongly recommended to define initially right after the Bromelia
    instantiation the set of Diameter messages to be used in each Diameter
    interface.
    
    Usually you create a :class:`Bromelia` instance in your main module.

    Usage::

        >>> from bromelia import Bromelia
        >>> app = Bromelia()
        >>> app.run()
    """

    def __init__(self, config_file=None):
        bromelia_logger.debug(f"Initializing Bromelia application")
        self.config_file = config_file
        self.configs = _convert_file_to_config(self.config_file, globals())
        self.app_name = get_app_name(self.config_file)
        self.routes = {}
        self._routes = {}
        self.sessions = {}
        self.g = Global()
        self.testing_answer = None
        
        self.recv_queues = None
        self.associations = None

        self.request_threshold = threading.Barrier(parties=REQUEST_THRESHOLD)
        self.request_id = 0

        self.answer_threshold = threading.Barrier(parties=ANSWER_THRESHOLD)
        self.answer_id = 0

        self.send_threshold = threading.Barrier(parties=SEND_THRESHOLD)


    def is_valid_session_key(self, session_key):
        session = self.sessions.get(session_key, None)
        if session is not None:
            return True
        return False


    def _create_applications(self, debug, is_logging):
        apps = list()
        
        for config in self.configs:
          app = Diameter(config=config, 
                         debug=debug, 
                         is_logging=is_logging,
                         app_name=self.app_name)
          apps.append(app)

        return apps


    def _run(self, debug, is_logging):
        apps = self._create_applications(debug, is_logging)

        bromelia_logger.debug(f"Route found: {self.routes})")

        with multiprocessing.Manager() as manager:
            for app in apps:
                bromelia_logger.debug(f"Initializing Worker for app {app}")
                worker = Worker(app, manager)

                bromelia_logger.debug(f"Starting Worker for app {app}")
                worker.start()

            self.recv_queues = Worker.recv_queues
            self.associations = Worker.associations

            bromelia_logger.debug(f"Loading recv_queues: {self.recv_queues}")
            bromelia_logger.debug(f"Loading associations: {self.associations}")

            self.main()


    def run(self, debug=False, is_logging=False, block=True):
        threading.Thread(target=self._run, 
                         args=(debug, is_logging)).start()

        #: Wait spinning up Diameter objects in the connection layer processes
        while True:
            time.sleep(BROMELIA_LOADING_TICKER)
            if self.associations is not None and self.recv_queues is not None:
                break

        #: Block til all Diameter objects are opened in the connection layer
        #: processes - that is, connection are established
        associations = copy(self.associations)
        while block:
            time.sleep(BROMELIA_LOADING_TICKER)
            if not associations:
                break

            for key in list(associations.keys()):
                association = associations[key]

                if association.is_open.is_set():
                    associations.pop(key)


    def main(self):
        bromelia_logger.debug(f"Starting Main Bromelia Loop")

        thrds = list()
        while True:
            time.sleep(BROMELIA_TICKER)

            for recv_queue, recv_lock in self.recv_queues:
                if not recv_queue.empty():
                    recv_lock.acquire()
                    message = recv_queue.get()
                    recv_lock.release()
                    
                    if message.header.is_request():
                        thrd = threading.Thread(name=f"recv_request_{self.request_id}", 
                                                target=self.callback_route, 
                                                args=(message,))
                        self.request_id += 1

                    else:
                        thrd = threading.Thread(name=f"recv_answer_{self.answer_id}", 
                                                target=self.handler_pending_answers, 
                                                args=(message,))
                        self.answer_id += 1

                    thrd.start()
                    thrds.append(thrd)

            for thrd in thrds:
                if not thrd.is_alive():
                    thrd.join()
                    thrds.remove(thrd)


    def route(self, application_id, command_code):
        def outer_function(route_function):
            def inner_function(*args, **kwargs):
                if application_id not in self.routes:
                    _route = {application_id: {command_code: route_function}}
                    self.routes.update(_route)
                else:
                    _command_code = {command_code: route_function}
                    self.routes[application_id].update(_command_code)

                self._routes.update({route_function.__name__: route_function})

            return inner_function()
        return outer_function


    def handler_pending_answers(self, message):
        try:
            self.answer_threshold.wait(timeout=PROCESS_TIMER)
        except threading.BrokenBarrierError:
            self.answer_threshold.reset()

        application_id = message.header.application_id
        hop_by_hop = message.header.hop_by_hop
        
        worker = self.associations[application_id]
        logging_info = f"[{worker.name}][{hop_by_hop.hex()}]"\
                       f"[PENDING:{len(worker.pending_answers)}]"

        bromelia_logger.debug(f"{logging_info} Check if it is an expected "\
                              f"answer")

        if hop_by_hop in worker.pending_answers.keys():
            bromelia_logger.debug(f"{logging_info} Found Hop-By-Hop in "\
                                  f"Pending answer")

            pending_answer = worker.pending_answers[hop_by_hop]

            pending_answer.message = message
            bromelia_logger.debug(f"{logging_info} Update Pending answer "\
                                  f"with the received answer")

            pending_answer.recv_event.set()
            pending_answer.stop_event.wait()

            worker.pending_answers.pop(message.header.hop_by_hop, None)


    def get_error_answer_for_request(self, request):
        application_id = request.header.application_id

        config = self.associations[application_id].app.config
        avps = [
                    SessionIdAVP(request.session_id_avp.data),
                    ResultCodeAVP(DIAMETER_UNABLE_TO_COMPLY),
                    OriginHostAVP(config["LOCAL_NODE_HOSTNAME"]),
                    OriginRealmAVP(config["LOCAL_NODE_REALM"]),
                    DestinationRealmAVP(request.origin_realm_avp.data),
                    DestinationHostAVP(request.origin_host_avp.data)
        ]

        return DiameterAnswer(header=request.header, avps=avps)


    def callback_route(self, request):
        try:
            self.request_threshold.wait(timeout=PROCESS_TIMER)
        except threading.BrokenBarrierError:
            self.request_threshold.reset()

        application_id = request.header.application_id
        command_code = request.header.command_code
        hop_by_hop = request.header.hop_by_hop

        worker = self.associations[application_id]
        logging_info = f"[{worker.name}][{hop_by_hop.hex()}]"

        callback_function = self.routes[application_id][command_code]
        bromelia_logger.debug(f"{logging_info} {callback_function}")

        try:
            answer = callback_function(request)

        except Exception as e:
            bromelia_logger.exception(f"{logging_info} Error has been "\
                                      f"raised in callback_function: {e.args}")
            answer = None

        if not isinstance(answer, DiameterAnswer):
            bromelia_logger.exception(f"{logging_info} There is no answer "\
                                      f"processed to be sent. We are sending "\
                                      f"UNABLE_TO_COMPLY")

            error_answer = self.get_error_answer_for_request(request)
            self.send_message(error_answer)

            raise BromeliaException("Route function must return "\
                                    "DiameterAnswer object")


        answer = get_formatted_answer_as_per_request(answer, request)

        if (is_3xxx_failure(answer) or 
            is_4xxx_failure(answer) or 
            is_5xxx_failure(answer)):

            answer.header.set_error_bit(True)

        if answer.has_avp("experimental_result_avp"):
            if answer.has_avp("result_code_avp"):
                answer.pop("result_code_avp")


        bromelia_logger.debug(f"{logging_info} Sending answer")
        self.send_message(answer)


    def send_message(self, message, recv_answer=True):
        application_id = message.header.application_id
        hop_by_hop = message.header.hop_by_hop

        if self.associations is None:
            return self.testing_answer
            
        worker = self.associations[application_id]
        logging_info = f"[{worker.name}][{hop_by_hop.hex()}]"

        bromelia_logger.debug(f"{logging_info} Application needs to send a "\
                              f"message")
        
        if not worker.is_running():
            bromelia_logger.debug(f"{logging_info} It seems the worker is "\
                                  f"not running anymore")
            return None

        bromelia_logger.debug(f"{logging_info} Putting message into "\
                              f"send_queue Queue")

        try:
            self.send_threshold.wait(timeout=SEND_THRESHOLD_TICKER)
        except threading.BrokenBarrierError:
            self.send_threshold.reset()

        worker.send_lock.acquire()
        worker.send_queue.put(message)
        worker.send_event.set()

        bromelia_logger.debug(f"{logging_info} Just put message into "\
                              f"send_queue Queue and notified send_event Event")

        if message.header.is_request() and recv_answer:
            pending_answer = PendingAnswer()
            worker.pending_answers.update({hop_by_hop: pending_answer})
            bromelia_logger.debug(f"{logging_info} Added Pending answer")

            pending_answer.recv_event.wait()
            bromelia_logger.debug(f"{logging_info} Notification from "\
                                  f"Pending answer")

            pending_answer.recv_event.clear()
            pending_answer.stop_event.set()
            bromelia_logger.debug(f"{logging_info} Cleanup of Pending answer")

            return pending_answer.message


    def load_messages_into_application_id(self, messages, application_id):
        def decorated_message(message):
            def proxy(**attrs):
                _config = None

                for config in self.configs:
                    for application in config["APPLICATIONS"]:
                        if application["app_id"] == application_id:
                            _config = config

                default_attrs = {
                            "origin_host": _config["LOCAL_NODE_HOSTNAME"],
                            "origin_realm": _config["LOCAL_NODE_REALM"],
                }

                attrs.update(**default_attrs)

                peer_node_realm = _config["PEER_NODE_REALM"]
                local_node_hostname = _config["LOCAL_NODE_HOSTNAME"]

                if issubclass(message, DiameterRequest):
                    request_attrs = {"destination_realm": peer_node_realm}
                    attrs.update(**request_attrs)

                if "session_id" in message.mandatory:
                    attrs.update({"session_id": local_node_hostname})

                return message(**attrs)

            return proxy

        application_string = get_application_string_by_id(application_id)
        self.__dict__.update({application_string: SimpleNamespace()})

        for message in messages:
            _name = [letter for letter in message.__name__ if letter.isupper()]
            short_message_name = "".join(_name).upper()

            setattr(self.__dict__[application_string], 
                    short_message_name, 
                    decorated_message(message))


    def get_route(self, route_name):
        return self._routes[route_name]


# -*- coding: utf-8 -*-
"""
    bromelia.lib.etsi_3gpp_s6a.messages
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    This module contains the Diameter protocol messages for 3GPP S6a/S6d 
    Application Id.

    :copyright: (c) 2020-present Henrique Marques Ribeiro.
    :license: MIT, see LICENSE for more details.
"""

import platform
import socket

from .avps import *

from ...base import DiameterRequest, DiameterAnswer
from ...constants import *


class AuthenticationInformationAnswer(DiameterAnswer):
    """Implementation of Authentication-Information-Answer (AIA) command as per 
    clause 7.2.6 of ETSI TS 129 272 V15.4.0 (2018-07).

    The Authentication-Information-Answer is indicated by the Command Code field
    set to 318 and Command Flag's 'R' bit cleared.

    Usage::

        >>> from bromelia.lib.etsi_3gpp_s6a import AIA
        >>> aia = AIA()
        >>> aia
        <Diameter Message: 318 [AIA] PXY, 16777251 [3GPP S6a], 5 AVP(s)>
    """

    mandatory = {
                    "session_id": SessionIdAVP,
                    "auth_session_state": AuthSessionStateAVP,
                    "origin_host": OriginHostAVP,
                    "origin_realm": OriginRealmAVP,
    }

    optionals = { 
                    # "drmp": DrmpAVP,
                    "vendor_specific_application_id": VendorSpecificApplicationIdAVP,
                    "result_code": ResultCodeAVP,
                    "experimental_result": ExperimentalResultAVP,
                    "error_diagnostic": ErrorDiagnosticAVP,
                    # "oc_supported_features": OcSupportedFeaturesAVP,
                    # "oc_olr": OcOlrAVP,
                    # "load": LoadAVP,
                    "supported_features": SupportedFeaturesAVP,
                    "authentication_info": AuthenticationInfoAVP,
                    "ue_usage_type": UeUsageTypeAVP,
                    "failed_avp": FailedAvpAVP,
                    "proxy_info": ProxyInfoAVP,
                    "route_record": RouteRecordAVP,
    }

    def __init__(self,
                 session_id=platform.node(),
                 drmp=None,
                 vendor_specific_application_id=[VendorIdAVP(VENDOR_ID_3GPP), AuthApplicationIdAVP(DIAMETER_APPLICATION_S6a_S6d)],
                 result_code=None,
                 experimental_result=None,
                 error_diagnostic=None,
                 auth_session_state=NO_STATE_MAINTAINED,
                 origin_host=platform.node(), 
                 origin_realm=socket.getfqdn(), 
                 oc_supported_features=None,
                 oc_olr=None,
                 load=None,
                 supported_features=None,
                 authentication_info=None,
                 ue_usage_type=None,
                 failed_avp=None,
                 proxy_info=None,
                 route_record=None,
                 **kwargs):

        DiameterAnswer.__init__(self, 
                                command_code=AUTHENTICATION_INFORMATION_MESSAGE, 
                                application_id=DIAMETER_APPLICATION_S6a_S6d)

        DiameterAnswer._load(self, locals())


class AuthenticationInformationRequest(DiameterRequest):
    """Implementation of Authentication-Information-Request (AIR) command as per 
    clause 7.2.5 of ETSI TS 129 272 V15.4.0 (2018-07).

    The Authentication-Information-Request is indicated by the Command Code 
    field set to 318 and the 'R' bit set in the Command Flags field.

    Usage::

        >>> from bromelia.lib.etsi_3gpp_s6a import AIR
        >>> air_avps = {
        ...     "destination_realm": "example.com",
        ...     "user_name": "frodo",
        ...     "visited_plmn_id": bytes.fromhex("ffffff")
        ... }
        >>> air = AIR(**air_avps)
        >>> air
        <Diameter Message: 318 [AIR] REQ|PXY, 16777251 [3GPP S6a], 8 AVP(s)>
    """    

    mandatory = {
                    "session_id": SessionIdAVP,
                    "auth_session_state": AuthSessionStateAVP,
                    "origin_host": OriginHostAVP,
                    "origin_realm": OriginRealmAVP,
                    "destination_realm": DestinationRealmAVP,
                    "user_name": UserNameAVP,
                    "visited_plmn_id": VisitedPlmnIdAVP,
    }

    optionals = {
                    # "drmp": DrmpAVP,
                    "vendor_specific_application_id": VendorSpecificApplicationIdAVP,
                    "destination_host": DestinationHostAVP,
                    # "oc_supported_features": OcSupportedFeaturesAVP,
                    "supported_features": SupportedFeaturesAVP,
                    "requested_eutran_authentication_info": RequestedEutranAuthenticationInfoAVP,
                    # "requested_utran_geran_authentication_info": RequestedUtranGeranAuthenticationInfoAVP,
                    "air_flags": AirFlagsAVP,
                    "proxy_info": ProxyInfoAVP,
                    "route_record": RouteRecordAVP,
    }

    def __init__(self, 
                 session_id=platform.node(), 
                 drmp=None,
                 vendor_specific_application_id=[VendorIdAVP(VENDOR_ID_3GPP), AuthApplicationIdAVP(DIAMETER_APPLICATION_S6a_S6d)],
                 auth_session_state=NO_STATE_MAINTAINED,
                 origin_host=platform.node(), 
                 origin_realm=socket.getfqdn(), 
                 destination_host=None,
                 destination_realm=None,
                 user_name=None,
                 oc_supported_features=None,
                 supported_features=None,
                 requested_eutran_authentication_info=None,
                 requested_utran_geran_authentication_info=None,
                 visited_plmn_id=None,
                 air_flags=None,
                 proxy_info=None,
                 route_record=None,
                 **kwargs):

        DiameterRequest.__init__(self, 
                                 command_code=AUTHENTICATION_INFORMATION_MESSAGE, 
                                 application_id=DIAMETER_APPLICATION_S6a_S6d)

        DiameterRequest._load(self, locals())


class CancelLocationAnswer(DiameterAnswer):
    """Implementation of Cancel-Location-Answer (CLA) command as per 
    clause 7.2.8 of ETSI TS 129 272 V15.4.0 (2018-07).

    The Cancel-Location-Answer is indicated by the Command Code field set to 
    317 and Command Flag's 'R' bit cleared.

    Usage::

        >>> from bromelia.lib.etsi_3gpp_s6a import CLA
        >>> cla = CLA()
        >>> cla
        <Diameter Message: 317 [CLA] PXY, 16777251 [3GPP S6a], 5 AVP(s)>
    """

    mandatory = {
                    "session_id": SessionIdAVP,
                    "auth_session_state": AuthSessionStateAVP,
                    "origin_host": OriginHostAVP,
                    "origin_realm": OriginRealmAVP,
    }

    optionals = { 
                    # "drmp": DrmpAVP,
                    "vendor_specific_application_id": VendorSpecificApplicationIdAVP,
                    "supported_features": SupportedFeaturesAVP,
                    "result_code": ResultCodeAVP,
                    "experimental_result": ExperimentalResultAVP,
                    "failed_avp": FailedAvpAVP,
                    "proxy_info": ProxyInfoAVP,
                    "route_record": RouteRecordAVP,
    }

    def __init__(self,
                 session_id=platform.node(),
                 drmp=None,
                 vendor_specific_application_id=[VendorIdAVP(VENDOR_ID_3GPP), AuthApplicationIdAVP(DIAMETER_APPLICATION_S6a_S6d)],
                 supported_features=None,
                 result_code=None,
                 experimental_result=None,
                 auth_session_state=NO_STATE_MAINTAINED,
                 origin_host=platform.node(), 
                 origin_realm=socket.getfqdn(), 
                 failed_avp=None,
                 proxy_info=None,
                 route_record=None,
                 **kwargs):

        DiameterAnswer.__init__(self, 
                                command_code=CANCEL_LOCATION_MESSAGE, 
                                application_id=DIAMETER_APPLICATION_S6a_S6d)

        DiameterAnswer._load(self, locals())


class CancelLocationRequest(DiameterRequest):
    """Implementation of Cancel-Location-Request (CLR) command as per 
    clause 7.2.7 of ETSI TS 129 272 V15.4.0 (2018-07).

    The Cancel-Location-Request is indicated by the Command Code field set to 
    317 and the 'R' bit set in the Command Flags field.

    Usage::

        >>> from bromelia.lib.etsi_3gpp_s6a import CLR
        >>> clr_avps = {
        ...     "destination_realm": "example.com",
        ...     "destination_host": "host.example.com",
        ...     "user_name": "frodo"
        ... }
        >>> clr = CLR(**clr_avps)
        >>> clr
        <Diameter Message: 317 [CLR] REQ|PXY, 16777251 [3GPP S6a], 9 AVP(s)>
    """    

    mandatory = {
                    "session_id": SessionIdAVP,
                    "auth_session_state": AuthSessionStateAVP,
                    "origin_host": OriginHostAVP,
                    "origin_realm": OriginRealmAVP,
                    "destination_host": DestinationHostAVP,
                    "destination_realm": DestinationRealmAVP,
                    "user_name": UserNameAVP,
                    "cancellation_type": CancellationTypeAVP,
    }

    optionals = {
                    # "drmp": DrmpAVP,
                    "vendor_specific_application_id": VendorSpecificApplicationIdAVP,
                    "supported_features": SupportedFeaturesAVP,
                    "clr_flags": ClrFlagsAVP,
                    "proxy_info": ProxyInfoAVP,
                    "route_record": RouteRecordAVP,
    }

    def __init__(self, 
                 session_id=platform.node(), 
                 drmp=None,
                 vendor_specific_application_id=[VendorIdAVP(VENDOR_ID_3GPP), AuthApplicationIdAVP(DIAMETER_APPLICATION_S6a_S6d)],
                 auth_session_state=NO_STATE_MAINTAINED,
                 origin_host=platform.node(), 
                 origin_realm=socket.getfqdn(), 
                 destination_host=None,
                 destination_realm=None,
                 user_name=None,
                 supported_features=None,
                 cancellation_type=CANCELLATION_TYPE_SUBSCRIPTION_WITHDRAWAL,
                 clr_flags=None,
                 proxy_info=None,
                 route_record=None,
                 **kwargs):

        DiameterRequest.__init__(self, 
                                 command_code=CANCEL_LOCATION_MESSAGE, 
                                 application_id=DIAMETER_APPLICATION_S6a_S6d)

        DiameterRequest._load(self, locals())


class NotifyAnswer(DiameterAnswer):
    """Implementation of Notify-Answer (NOA) command as per clause 7.2.18 of
    ETSI TS 129 272 V15.4.0 (2018-07).

    The Notify-Answer is indicated by the Command Code field set to 323 and
    Command Flag's 'R' bit cleared.

    Usage::

        >>> from bromelia.lib.etsi_3gpp_s6a import NOA
        >>> noa = NOA()
        >>> noa
        <Diameter Message: 323 [NOA] PXY, 16777251 [3GPP S6a], 5 AVP(s)>
    """

    mandatory = {
                    "session_id": SessionIdAVP,
                    "auth_session_state": AuthSessionStateAVP,
                    "origin_host": OriginHostAVP,
                    "origin_realm": OriginRealmAVP,
    }

    optionals = { 
                    # "drmp": DrmpAVP,
                    "vendor_specific_application_id": VendorSpecificApplicationIdAVP,
                    "result_code": ResultCodeAVP,
                    "experimental_result": ExperimentalResultAVP,
                    # "oc_supported_features": OcSupportedFeaturesAVP,
                    # "oc_olr": OcOlrAVP,
                    # "load": LoadAVP,
                    "supported_features": SupportedFeaturesAVP,
                    "failed_avp": FailedAvpAVP,
                    "proxy_info": ProxyInfoAVP,
                    "route_record": RouteRecordAVP,
    }

    def __init__(self,
                 session_id=platform.node(),
                 drmp=None,
                 vendor_specific_application_id=[VendorIdAVP(VENDOR_ID_3GPP), AuthApplicationIdAVP(DIAMETER_APPLICATION_S6a_S6d)],
                 result_code=None,
                 experimental_result=None,
                 auth_session_state=NO_STATE_MAINTAINED,
                 origin_host=platform.node(), 
                 origin_realm=socket.getfqdn(), 
                 oc_supported_features=None,
                 oc_olr=None,
                 load=None,
                 supported_features=None,
                 failed_avp=None,
                 proxy_info=None,
                 route_record=None,
                 **kwargs):

        DiameterAnswer.__init__(self, 
                                command_code=NOTIFY_MESSAGE, 
                                application_id=DIAMETER_APPLICATION_S6a_S6d)

        DiameterAnswer._load(self, locals())


class NotifyRequest(DiameterRequest):
    """Implementation of Notify-Request (NOR) command as per clause 7.2.17 of 
    ETSI TS 129 272 V15.4.0 (2018-07).

    The Notify-Request is indicated by the Command Code field set to 323 and 
    the 'R' bit set in the Command Flags field.

    Usage::

        >>> from bromelia.lib.etsi_3gpp_s6a import NOR
        >>> nor_avps = {
        ...     "destination_realm": "example.com",
        ...     "user_name": "frodo",
        ... }
        >>> nor = NOR(**nor_avps)
        >>> nor
        <Diameter Message: 323 [NOR] REQ|PXY, 16777251 [3GPP S6a], 8 AVP(s)>
    """    

    mandatory = {
                    "session_id": SessionIdAVP,
                    "auth_session_state": AuthSessionStateAVP,
                    "origin_host": OriginHostAVP,
                    "origin_realm": OriginRealmAVP,
                    "destination_realm": DestinationRealmAVP,
                    "user_name": UserNameAVP,
    }

    optionals = {
                    "vendor_specific_application_id": VendorSpecificApplicationIdAVP,
                    # "drmp": DrmpAVP,
                    "destination_host": DestinationHostAVP,
                    # "oc_supported_features": OcSupportedFeaturesAVP,
                    "supported_features": SupportedFeaturesAVP,
                    "terminal_information": TerminalInformationAVP,
                    "mip6_agent_info": Mip6AgentInfoAVP,
                    "visited_network_identifier": VisitedNetworkIdentifierAVP,
                    "context_identifier": ContextIdentifierAVP,
                    "service_selection": ServiceSelectionAVP,
                    "alert_reason": AlertReasonAVP,
                    "ue_srvcc_capability": UeSrvccCapabilityAVP,
                    "nor_flags": NorFlagsAVP,
                    "homogeneous_support_of_ims_voice_over_ps_sessions": HomogeneousSupportOfImsVoiceOverPsSessionsAVP,
                    # "maximum_ue_availability_time": MaximumUeAvailabilityTimeAVP,
                    # "monitoring_event_config_status": MonitoringEventConfigStatusAVP,
                    # "emergency_services": EmergencyServicesAVP,
                    "proxy_info": ProxyInfoAVP,
                    "route_record": RouteRecordAVP,
    }

    def __init__(self, 
                 session_id=platform.node(), 
                 vendor_specific_application_id=[VendorIdAVP(VENDOR_ID_3GPP), AuthApplicationIdAVP(DIAMETER_APPLICATION_S6a_S6d)],
                 drmp=None,
                 auth_session_state=NO_STATE_MAINTAINED,
                 origin_host=platform.node(), 
                 origin_realm=socket.getfqdn(), 
                 destination_host=None,
                 destination_realm=None,
                 user_name=None,
                 oc_supported_features=None,
                 supported_features=None,
                 terminal_information=None,
                 mip6_agent_info=None,
                 visited_network_identifier=None,
                 context_identifier=None,
                 service_selection=None,
                 alert_reason=None,
                 ue_srvcc_capability=None,
                 nor_flags=None,
                 homogeneous_support_of_ims_voice_over_ps_sessions=None,
                 maximum_ue_availability_time=None,
                 monitoring_event_config_status=None,
                 emergency_services=None,
                 proxy_info=None,
                 route_record=None,
                 **kwargs):

        DiameterRequest.__init__(self, 
                                 command_code=NOTIFY_MESSAGE, 
                                 application_id=DIAMETER_APPLICATION_S6a_S6d)

        DiameterRequest._load(self, locals())


class PurgeUeAnswer(DiameterAnswer):
    """Implementation of Purge-UE-Answer (PUA) command as per clause 7.2.14 of
    ETSI TS 129 272 V15.4.0 (2018-07).

    The Purge-UE-Answer is indicated by the Command Code field set to 321 and
    Command Flag's 'R' bit cleared.

    Usage::

        >>> from bromelia.lib.etsi_3gpp_s6a import PUA
        >>> pua = PUA()
        >>> pua
        <Diameter Message: 321 [PUA] PXY, 16777251 [3GPP S6a], 5 AVP(s)>
    """

    mandatory = {
                    "session_id": SessionIdAVP,
                    "auth_session_state": AuthSessionStateAVP,
                    "origin_host": OriginHostAVP,
                    "origin_realm": OriginRealmAVP,
    }

    optionals = { 
                    # "drmp": DrmpAVP,
                    "vendor_specific_application_id": VendorSpecificApplicationIdAVP,
                    "result_code": ResultCodeAVP,
                    "experimental_result": ExperimentalResultAVP,
                    # "oc_supported_features": OcSupportedFeaturesAVP,
                    # "oc_olr": OcOlrAVP,
                    # "load": LoadAVP,
                    "pua_flags": PuaFlagsAVP,
                    "failed_avp": FailedAvpAVP,
                    "proxy_info": ProxyInfoAVP,
                    "route_record": RouteRecordAVP,
    }

    def __init__(self,
                 session_id=platform.node(),
                 drmp=None,
                 vendor_specific_application_id=[VendorIdAVP(VENDOR_ID_3GPP), AuthApplicationIdAVP(DIAMETER_APPLICATION_S6a_S6d)],
                 result_code=None,
                 experimental_result=None,
                 auth_session_state=NO_STATE_MAINTAINED,
                 origin_host=platform.node(), 
                 origin_realm=socket.getfqdn(), 
                 oc_supported_features=None,
                 oc_olr=None,
                 load=None,
                 pua_flags=None,
                 failed_avp=None,
                 proxy_info=None,
                 route_record=None,
                 **kwargs):

        DiameterAnswer.__init__(self, 
                                command_code=PURGE_UE_MESSAGE, 
                                application_id=DIAMETER_APPLICATION_S6a_S6d)

        DiameterAnswer._load(self, locals())


class PurgeUeRequest(DiameterRequest):
    """Implementation of Purge-UE-Request (PUR) command as per clause 7.2.13 of
    ETSI TS 129 272 V15.4.0 (2018-07).

    The Purge-UE-Request is indicated by the Command Code field set to 321 and
    the 'R' bit set in the Command Flags field.

    Usage::

        >>> from bromelia.lib.etsi_3gpp_s6a import PUR
        >>> pur_avps = {
        ...     "destination_realm": "example.com",
        ...     "user_name": "frodo",
        ... }
        >>> pur = PUR(**pur_avps)
        >>> pur
        <Diameter Message: 321 [PUR] REQ|PXY, 16777251 [3GPP S6a], 8 AVP(s)>
    """    

    mandatory = {
                    "session_id": SessionIdAVP,
                    "auth_session_state": AuthSessionStateAVP,
                    "origin_host": OriginHostAVP,
                    "origin_realm": OriginRealmAVP,
                    "destination_realm": DestinationRealmAVP,
                    "user_name": UserNameAVP,
    }

    optionals = {
                    # "drmp": DrmpAVP,
                    "vendor_specific_application_id": VendorSpecificApplicationIdAVP,
                    "destination_host": DestinationHostAVP,
                    # "oc_supported_features": OcSupportedFeaturesAVP,
                    "pur_flags": PurFlagsAVP,
                    "supported_features": SupportedFeaturesAVP,
                    # "eps_location_information": EpsLocationInformationAVP,
                    "proxy_info": ProxyInfoAVP,
                    "route_record": RouteRecordAVP,
    }

    def __init__(self, 
                 session_id=platform.node(), 
                 drmp=None,
                 vendor_specific_application_id=[VendorIdAVP(VENDOR_ID_3GPP), AuthApplicationIdAVP(DIAMETER_APPLICATION_S6a_S6d)],
                 auth_session_state=NO_STATE_MAINTAINED,
                 origin_host=platform.node(), 
                 origin_realm=socket.getfqdn(), 
                 destination_host=None,
                 destination_realm=None,
                 user_name=None,
                 oc_supported_features=None,
                 pur_flags=None,
                 supported_features=None,
                 eps_location_information=None,
                 proxy_info=None,
                 route_record=None,
                 **kwargs):

        DiameterRequest.__init__(self, 
                                 command_code=PURGE_UE_MESSAGE, 
                                 application_id=DIAMETER_APPLICATION_S6a_S6d)

        DiameterRequest._load(self, locals())


class UpdateLocationAnswer(DiameterAnswer):
    """Implementation of Update-Location-Answer (ULA) command as per 
    clause 7.2.4 of ETSI TS 129 272 V15.4.0 (2018-07).

    The Update-Location-Answer is indicated by the Command Code field
    set to 316 and Command Flag's 'R' bit cleared.

    Usage::

        >>> from bromelia.lib.etsi_3gpp_s6a import ULA
        >>> ula = ULA()
        >>> ula
        <Diameter Message: 316 [ULA] PXY, 16777251 [3GPP S6a], 5 AVP(s)>
    """

    mandatory = {
                    "session_id": SessionIdAVP,
                    "auth_session_state": AuthSessionStateAVP,
                    "origin_host": OriginHostAVP,
                    "origin_realm": OriginRealmAVP,
    }

    optionals = { 
                    # "drmp": DrmpAVP,
                    "vendor_specific_application_id": VendorSpecificApplicationIdAVP,
                    "result_code": ResultCodeAVP,
                    "experimental_result": ExperimentalResultAVP,
                    "error_diagnostic": ErrorDiagnosticAVP,
                    # "oc_supported_features": OcSupportedFeaturesAVP,
                    # "oc_olr": OcOlrAVP,
                    # "load": LoadAVP,
                    "supported_features": SupportedFeaturesAVP,
                    "ula_flags": UlaFlagsAVP,
                    "subscription_data": SubscriptionDataAVP,
                    # "reset_id": ResetIdAVP,
                    "failed_avp": FailedAvpAVP,
                    "proxy_info": ProxyInfoAVP,
                    "route_record": RouteRecordAVP,
    }

    def __init__(self,
                 session_id=platform.node(),
                 drmp=None,
                 vendor_specific_application_id=[VendorIdAVP(VENDOR_ID_3GPP), AuthApplicationIdAVP(DIAMETER_APPLICATION_S6a_S6d)],
                 result_code=None,
                 experimental_result=None,
                 error_diagnostic=None,
                 auth_session_state=NO_STATE_MAINTAINED,
                 origin_host=platform.node(), 
                 origin_realm=socket.getfqdn(), 
                 oc_supported_features=None,
                 oc_olr=None,
                 load=None,
                 supported_features=None,
                 ula_flags=None,
                 subscription_data=None,
                 reset_id=None,
                 failed_avp=None,
                 proxy_info=None,
                 route_record=None,
                 **kwargs):

        DiameterAnswer.__init__(self, 
                                command_code=UPDATE_LOCATION_MESSAGE, 
                                application_id=DIAMETER_APPLICATION_S6a_S6d)

        DiameterAnswer._load(self, locals())


class UpdateLocationRequest(DiameterRequest):
    """Implementation of Update-Location-Request (ULR) command as per 
    clause 7.2.3 of ETSI TS 129 272 V15.4.0 (2018-07).

    The Update-Location-Request is indicated by the Command Code field set to 
    316 and the 'R' bit set in the Command Flags field.

    Usage::

        >>> from bromelia.lib.etsi_3gpp_s6a import ULR
        >>> ulr_avps = {
        ...     "destination_realm": "example.com",
        ...     "user_name": "frodo",
        ...     "visited_plmn_id": bytes.fromhex("ffffff")
        ... }
        >>> ulr = ULR(**ulr_avps)
        >>> ulr
        <Diameter Message: 316 [ULR] REQ|PXY, 16777251 [3GPP S6a], 10 AVP(s)>
    """    

    mandatory = {
                    "session_id": SessionIdAVP,
                    "auth_session_state": AuthSessionStateAVP,
                    "origin_host": OriginHostAVP,
                    "origin_realm": OriginRealmAVP,
                    "destination_realm": DestinationRealmAVP,
                    "user_name": UserNameAVP,
                    "rat_type": RatTypeAVP,
                    "ulr_flags": UlrFlagsAVP,
                    "visited_plmn_id": VisitedPlmnIdAVP,
    }

    optionals = {
                    # "drmp": DrmpAVP,
                    "vendor_specific_application_id": VendorSpecificApplicationIdAVP,
                    "destination_host": DestinationHostAVP,
                    # "oc_supported_features": OcSupportedFeaturesAVP,
                    "supported_features": SupportedFeaturesAVP,
                    "terminal_information": TerminalInformationAVP,
                    "ue_srvcc_capability": UeSrvccCapabilityAVP,
                    # "sgsn_number": SgsnNumberAVP,
                    "homogeneous_support_of_ims_voice_over_ps_sessions": HomogeneousSupportOfImsVoiceOverPsSessionsAVP,
                    # "gmlc_address": GmlcAddressAVP,
                    # "active_apn": ActiveApnAVP,
                    # "equivalent_plmn_list": EquivalentPlmnListAVP,
                    # "mme_number_for_mt_sms": MmeNumberForMtSmsAVP,
                    # "sms_register_request": SmsRegisterRequestAVP,
                    # "sgs_mme_identity": SgsMmeIdentityAVP,
                    # "copled_node_diameter_id": CopledNodeDiameterIdAVP,
                    # "adjacent_plmns": AdjacentPlmnsAVP,
                    "supported_services": SupportedServicesAVP,
                    "proxy_info": ProxyInfoAVP,
                    "route_record": RouteRecordAVP,
    }

    def __init__(self, 
                 session_id=platform.node(), 
                 drmp=None,
                 vendor_specific_application_id=[VendorIdAVP(VENDOR_ID_3GPP), AuthApplicationIdAVP(DIAMETER_APPLICATION_S6a_S6d)],
                 auth_session_state=NO_STATE_MAINTAINED,
                 origin_host=platform.node(), 
                 origin_realm=socket.getfqdn(), 
                 destination_host=None,
                 destination_realm=None,
                 user_name=None,
                 oc_supported_features=None,
                 supported_features=None,
                 terminal_information=None,
                 rat_type=RAT_TYPE_EUTRAN,
                 ulr_flags=3,
                 visited_plmn_id=None,
                 ue_srvcc_capability=None,
                 sgsn_number=None,
                 homogeneous_support_of_ims_voice_over_ps_sessions=None,
                 gmlc_address=None,
                 active_apn=None,
                 equivalent_plmn_list=None,
                 mme_number_for_mt_sms=None,
                 sms_register_request=None,
                 sgs_mme_identity=None,
                 copled_node_diameter_id=None,
                 adjacent_plmns=None,
                 supported_services=None,
                 proxy_info=None,
                 route_record=None,
                 **kwargs):

        DiameterRequest.__init__(self, 
                                 command_code=UPDATE_LOCATION_MESSAGE, 
                                 application_id=DIAMETER_APPLICATION_S6a_S6d)

        DiameterRequest._load(self, locals())